/**************************************************************************************************
  MyTimers.c

  Copyright (C) 2015  Val Gretchev

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This is a compendium of hardware and software timers. Use what you need and ignore the rest.

**************************************************************************************************/

// Include Files
#include <stddef.h>
#include "FreeRTOSConfig.h"
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"
#include "timers.h"
#include "projdefs.h"
#include "portmacro.h"

#include "MyTimers.h"
#include "Buffers.h"
#include "USB_Driver3.h"


sbit Ind at GPIOD_ODR.B8; // PD8 output name


// System counters
int TimerFlag = 0;                // Set by SysTick interrupt handler
unsigned long SysTimeCnt;         // Keeps incrementing wrapping to 0 at max count
unsigned int Msec = 0;            // Counts milliseconds from 0 to 999
unsigned int DelayCntr[2] = {0};  // Debounce counter unused
char CntrPtr;

// Hardware timers
unsigned Ratio;
char Timer5Flag = 0;
unsigned RatioBuz;

// Handle for software timer
TimerHandle_t xOneShotTimer;
TimerHandle_t xAutoReloadTimer;

// Variables
struct SoftwareTimers TimerPool[NUMTIMERS];

/*************************************************************************************************/
// Hardware Level Functions
/*************************************************************************************************/

/**************************************************************************************************
  System Timer Interrupt Handler
  NOTE: can't use SysTick with FreeRTOS - already allocated to FreeRTOS. Use Timer 11 instead.
**************************************************************************************************/
/*
void SystemTimerInterrupt() iv IVT_INT_SysTick ics ICS_AUTO {
     if (COUNTFLAG_bit) {              // Interrupts every millisecond
         SysTimeCnt++;

         // For each delay counter if value > 0 subtract 1
         for (CntrPtr = 0; CntrPtr < 2; CntrPtr++) {
              if (DelayCntr[CntrPtr] > 0)
                  DelayCntr[CntrPtr]--;
         }

         Msec += 1;                    // Increment millisecond counter
         if (Msec > 999) {             // Count to 1000
             Msec = 0;                 // Set it to 0
             TimerFlag = 1;            // Set 1 second flag - disabled
         }
         irqhandler_timer();           // Service timer pool
     }
}
*/
/**************************************************************************************************
  Initialize system timer
  NOTE: can't use SysTick with FreeRTOS - already allocated to FreeRTOS
**************************************************************************************************/
/*
void Timer_Init() {
     init_timerpool();        // Make all timers free

     STK_LOAD = 21000; // Set reload value register for 168 MHz
     STK_VAL = 0; // Reset timer counter
     ENABLE_STK_CTRL_bit = 1; // Enable timer

     TICKINT_bit = 1; // Enable timer interrupt

     USB_PutQueue("Timer initialized\r\n");
}
*/

void Timer_Initialize(void) {
     init_timerpool();        // Make all timers free
     Timer11_Initialize();    // This timer is used instead of SysTick for software timers
     Timer11_Start();
}

/**************************************************************************************************
  Timer processing is done if timer flag is set
**************************************************************************************************/
void TimerCheck() {
     if (TimerFlag == 1) {
         TimerFlag = 0;                 // Reset 1 second flag
         //USB_PutQueue("Tick\r\n");      // Diagnostic
         IndTog();                      // Diagnostic
     }
}

//-------------------------------------------------------------------------------------------------



/*************************************************************************************************/
// Software Level Functions
/*************************************************************************************************/

/**************************************************************************************************
File:         Timers.c
Description:  Implements simple timer pool. No timer chaining
Resolution:   +0/-0.001 second

Author: Jari Lahti (jari.lahti@violasystems.com)
Version Info: 22.1.2004 - First version (JaL)
**************************************************************************************************/



/**************************************************************************************************
Function:   init_timerpool
Parameters: none
Return val: void
Desc:       Initialize timer pool
**************************************************************************************************/
void init_timerpool(void) {
     unsigned short int i;

     for (i=0; i < NUMTIMERS; i++) {
          TimerPool[i].TimerField = 0;
          TimerPool[i].Free = TRUE;
     }
}

/**************************************************************************************************
Function:   get_timer
Parameters: none
Return val: unsigned short int - returns handle of free timer
Desc:       Gets free timer from the timer pool
**************************************************************************************************/
unsigned short int get_timer(void) {
     unsigned short int i;

     for (i=0; i < NUMTIMERS; i++) {
          if (TimerPool[i].Free == TRUE) {  // If a free timer is found
              TimerPool[i].Free = FALSE;    // Mark it reserved
              return i;                     // Return timer handle
          }
     }

     // Error Check
     USB_PutQueue("TIMERS: No Timers, Resetting..\r\n");
     SystemReset();
     return (0);
}

/**************************************************************************************************
Function:   free_timer
Parameters: none
Return val: unsigned short int nbr - number
Desc:       Puts the given timer back to the timer pool
**************************************************************************************************/
void free_timer(unsigned short int nbr) {
     if (nbr > (NUMTIMERS-1))          // Make a simple check
         return;

     TimerPool[nbr].Free = TRUE;
}

/**************************************************************************************************
Function:   set_timer
Parameters: unsigned short int nbr - number
            unsigned long int tout - time out value
Return val: none
Desc:       Initializes given timer to given time-out
**************************************************************************************************/
void set_timer(unsigned short int nbr, unsigned long int tout) {
     if (nbr > (NUMTIMERS-1))          // Make a simple check
         return;

     if (TimerPool[nbr].Free == TRUE)
         return;

     TimerPool[nbr].TimerField = tout;
}

/**************************************************************************************************
Function:   check_timer
Parameters: unsigned short int nbr - number
Return val: unsigned long int - returns the value of given timer
Desc:       Return the value of given timer
**************************************************************************************************/
unsigned long int check_timer(unsigned short int nbr) {
     return TimerPool[nbr].TimerField;
}

/**************************************************************************************************
Function:   irqhandler_timer
Parameters: void
Return val: void
Desc:       Call this function on timer interrupt
**************************************************************************************************/
void irqhandler_timer(void) {
     unsigned short int i;

     for (i=0; i<NUMTIMERS; i++) {
          if (TimerPool[i].Free == FALSE && TimerPool[i].TimerField != 0)
          TimerPool[i].TimerField--;
     }
}
//-------------------------------------------------------------------------------------------------






/*************************************************************************************************/

// Hardware timers

/**************************************************************************************************
  Advanced-control Timers (TIM1)
  Timer 1 Channel 1 on PE9 is used for PWM control of BLED
  Timer1 is on APB2 which runs at 60MHz at the CPU frequency of 120MHz.
**************************************************************************************************/
void Timer1_Initialize(void) {

     //GPIO_Digital_Output(&GPIOE_BASE, _GPIO_PINMASK_9);

     // Use library to set things up then access timer registers directly
     Ratio = PWM_TIM1_Init(50);  // Set frequency to 50Hz or period to 20ms
     PWM_TIM1_Set_Duty(Ratio/2, _PWM_NON_INVERTED, _PWM_CHANNEL1);  // Set duty cycle to one-half of period
     PWM_TIM1_Start(_PWM_CHANNEL1, &_GPIO_MODULE_TIM1_CH1_PE9);

     // The PWM frequency is 60000000/50/24000 = 50Hz or 20ms
     TIM1_PSC = 49;          // Set the prescale register to divide by 49+1=50
     TIM1_ARR = 24000;       // Set the auto-load register to divide by 24000

     // The desired PWM pulse period in ms is PP/20*24000
     // The center point for the PWM is 10ms
     TIM1_CCR1 = 12000;      // The count to be loaded into TIM1_CCR1 is 10/20*24000=12000
     Ratio = TIM1_CCR1;      // Replace library value with calculated value
     CEN_bit = 1;            // Enable counter
}

/**************************************************************************************************
  Increase PWM on-time (pulse period)
**************************************************************************************************/
void PwmUp(void) {
     char *BufNum;

     Ratio = TIM1_CCR1;
     if (Ratio < 24000) {
         Ratio += 480;
         TIM1_CCR1 = Ratio;
     }
     // Diagnostic
     BufNum = GetBuffer();
     sprintf(BufNum,"Duty: %d\r\n",Ratio);
     USB_PutQueue(BufNum);
}

/**************************************************************************************************
  Decrease PWM on-time (pulse period)
**************************************************************************************************/
void PwmDn(void) {
     char *BufNum;

     Ratio = TIM1_CCR1;
     if (Ratio > 0) {
         Ratio -= 480;
         TIM1_CCR1 = Ratio;
     }
     // Diagnostic
     BufNum = GetBuffer();
     sprintf(BufNum,"Duty: %d\r\n",Ratio);
     USB_PutQueue(BufNum);
}

/**************************************************************************************************
  BLED PWM fully on
**************************************************************************************************/
void PwmMaxOn(void) {
     char *BufNum;

     Ratio = 24000;
     TIM1_CCR1 = Ratio;

     // Diagnostic
     BufNum = GetBuffer();
     sprintf(BufNum,"Duty: %d\r\n",Ratio);
     USB_PutQueue(BufNum);
}

/**************************************************************************************************
  BLED PWM fully off
**************************************************************************************************/
void PwmMaxOff(void) {
     char *BufNum;

     Ratio = 0;
     TIM1_CCR1 = Ratio;

     // Diagnostic
     BufNum = GetBuffer();
     sprintf(BufNum,"Duty: %d\r\n",Ratio);
     USB_PutQueue(BufNum);
}
//-------------------------------------------------------------------------------------------------


/**************************************************************************************************
  Timer3 Interrupt Handler
**************************************************************************************************/
void Timer3Interrupt() iv IVT_INT_TIM3 ics ICS_AUTO {
     UIF_TIM3_SR_bit = 0;
     IndTog();
}

/**************************************************************************************************
  Timer3 Initialization
  APB1 timer clock runs at 60MHz. Timer 3 channel 1 is on pin PC6.
  This timer can be used as a diagnostic aid. Put the Timer3_Ch1_Start() function call anywhere
  in your code to flash PC6 LED.
  Set prescaler to divide by 6000: 60000000 / 6000 = 10000 Hz
  Set auto-reload to divide by 10000: 10000 / 10000 = 1 Hz
  Therefore interrupt will occur once per second
**************************************************************************************************/
void Timer3_Initialize(void) {

     /* Set alternate-function to output the timer on PC6
     GPIOCEN_bit = 1;             // Set GPIOD clock enable bit
     GPIOC_MODER &= 0xFFFFCFFF;   // Clear port C MODER-6 bits
     GPIOC_MODER |= 0x00002000;   // Set port C MODER-6 bits to 10: alternate function mode
     GPIOC_AFRL  &= 0xF0FFFFFF;   // Clear port C AFRH-6 bits
     GPIOC_AFRL  |= 0x02000000;   // Set port C AFRH-6 bits to 0010: AF2 TIM3_CH1
     */

     // Set interrupt vector
     NVIC_IntEnable(IVT_INT_TIM3);// Allow interrupts in NVIC

     // RCC_APB1ENR
     TIM3EN_bit = 1;               // Enable TIM3 clock in RCC_APB1ENR.B1

     //TIM3_CR1  = 0x0008;          // Set OPM bit (one-pulse mode)
     //OPM_TIM3_CR1_bit = 1;        // Use this instead

     //TIM3_PSC  = 99;              // Prescaler value
     TIM3_PSC  = 6000;            // Prescaler value
     TIM3_ARR  = 10000;           // Pulse width value
     TIM3_CCR1 = 10;              // Delay value

     TIM3_DIER = 0x0001;          // Update interrupt enable
     UIE_TIM3_DIER_bit = 1;       // Set update interrupt enable
     URS_TIM3_CR1_bit = 0;        // Clear update request source
     //UG_TIM3_EGR_bit = 1;         // Generate update

     //TIM3_CCMR1_Output = 0x0070;  // PWM mode 2 for CH1
     //TIM3_CCER = 0x0001;          // Channels 1 is connected to output

     USB_PutQueue("Timer3 initialized\r\n");
}
/**************************************************************************************************
  Start Timer3 Channel1
**************************************************************************************************/
void Timer3_Ch1_Start(void) {

     CEN_TIM3_CR1_bit = 1;         // Enable timer 3 channel 1

     USB_PutQueue("Timer3 started\r\n");
}

/**************************************************************************************************
  Stop Timer3 Channel1
**************************************************************************************************/
void Timer3_Ch1_Stop(void) {

     CEN_TIM3_CR1_bit = 0;         // Disable timer 3 channel 1

     USB_PutQueue("Timer3 stopped\r\n");
}

/**************************************************************************************************
  Generate update on Timer3 Channel1
**************************************************************************************************/
void Timer3_Update(void) {
     UG_TIM3_EGR_bit = 1;         // Generate update
     USB_PutQueue("Timer3 interrupt generated\r\n");
}
//-------------------------------------------------------------------------------------------------




/**************************************************************************************************
  Timer4 Interrupt Handler
**************************************************************************************************/
void Timer4Interrupt() iv IVT_INT_TIM4 ics ICS_AUTO {
     UIF_TIM4_SR_bit = 0;
     IndTog();
}


/**************************************************************************************************
  Timer4 Initialization
  APB1 timer clock runs at 60MHz. Timer 4 channel 3 is on pin PD14.
  This timer can be used as a diagnostic aid. Put the Timer4_Ch1_Start() function call anywhere
  in your code to flash PD14 LED.
**************************************************************************************************/
void Timer4_Initialize(void) {
     GPIODEN_bit = 1;             // Set GPIOD clock enable bit
     //GPIOD_MODER &= 0xFCFFFFFF;   // Clear port D MODER-12 bits
     //GPIOD_MODER &= 0xF3FFFFFF;   // Clear port D MODER-13 bits
     GPIOD_MODER &= 0xCFFFFFFF;   // Clear port D MODER-14 bits
     //GPIOD_MODER |= 0x02000000;   // Set port D MODER-12 to 10: alternate function mode
     //GPIOD_MODER |= 0x08000000;   // Set port D MODER-13 to 10: alternate function mode
     GPIOD_MODER |= 0x20000000;   // Set port D MODER-14 to 10: alternate function mode
     //GPIOD_AFRH  &= 0xFFF0FFFF;   // Clear port D AFRH-12 bits
     //GPIOD_AFRH  &= 0xFF0FFFFF;   // Clear port D AFRH-13 bits
     GPIOD_AFRH  &= 0xF0FFFFFF;   // Clear port D AFRH-14 bits
     //GPIOD_AFRH  |= 0x00020000;   // Set port d AFRH-12 to 0010: AF2 TIM4_CH1
     //GPIOD_AFRH  |= 0x00200000;   // Set port d AFRH-13 to 0010: AF2 TIM4_CH2
     GPIOD_AFRH  |= 0x02000000;   // Set port d AFRH-14 to 0010: AF2 TIM4_CH3



     // Set interrupt vector
     NVIC_IntEnable(IVT_INT_TIM4);// Allow interrupts in NVIC

     // RCC_APB1ENR
     TIM4EN_bit = 1;               // Enable TIM4 clock in RCC_APB1ENR.B2

     TIM4_CR1  = 0x0008;          // Set OPM bit (one-pulse mode)
     TIM4_ARR  = 10000;           // Pulse width value
     //TIM4_CCR1 = 10;              // Pulse delay value for CH1
     //TIM4_CCR2 = 10;              // Pulse delay value for CH2
     TIM4_CCR3 = 10;              // Pulse delay value for CH3
     TIM4_PSC  = 99;              // Prescaler value

     UG_TIM4_EGR_bit = 1;         // Generate update

     //TIM4_CCMR1_Output = 0x0070;  // PWM mode 2 for CH1
     //TIM4_CCMR1_Output = 0x7000;  // PWM mode 2 for CH2
     TIM4_CCMR2_Output = 0x0070;  // PWM mode 2 for CH3

     //TIM4_CCER = 0x0001;          // Channels 1 is connected to output
     //TIM4_CCER = 0x0010;          // Channels 2 is connected to output
     TIM4_CCER = 0x0100;          // Channels 3 is connected to output

     USB_PutQueue("Timer4 initialized\r\n");
}
/**************************************************************************************************
  Start Timer 4 Channel 3
**************************************************************************************************/
void Timer4_Ch1_Start(void) {
     CEN_TIM4_CR1_bit = 1;         // Enable timer 4 channel 3
     USB_PutQueue("Timer4 started\r\n");
}
//-------------------------------------------------------------------------------------------------


/**************************************************************************************************
  Timer5 Interrupt Handler
**************************************************************************************************/
void Timer5_interrupt() iv IVT_INT_TIM5 {
     UIF_TIM5_SR_bit = 0;
     Timer5Flag = 1;
     //AD_IntFlag = 1;    // Simulate interrupts from ADS1246
     //USB_PutQueue("Timer5 tick\r\n"); // Diagnostic when timer5 is set for 1 second interval
     // NOTE: can't use USB_PutQueue from interrupt when using FreeRTOS - hangs up.
     IndTog();
}

/**************************************************************************************************
  Timer5 interrupt initialization code
  Timer5 is on APB1 clock bus at 60 MHz.
  Set prescaler to divide by 6000: 60000000 / 6000 = 10000 Hz
  Set auto-reload to divide by 10000: 10000 / 10000 = 1 Hz
  Therefore interrupt will occur once per second
**************************************************************************************************/
void Timer5_Initialize(void) {
     TIM5EN_bit = 1;                   // Enable TIM5 clock in RCC_APB1ENR
     CEN_TIM5_CR1_bit = 0;             // Disable timer 5 while setting registers
     // Set timer5 for 1 second interval
     TIM5_PSC = 5999;                  // Prescaler divide by 6000
     TIM5_ARR = 10000;                 // Reload value to 10000
     NVIC_IntEnable(IVT_INT_TIM5);     // Set timer 5 interrupt vector
     UIE_TIM5_DIER_bit = 1;            // Enable timer5 update interrupts
     //TIM5_CR1.CEN = 1;                 // Start timer5 (do it in Timer5Start function)
     USB_PutQueue("Timer5 initialized\r\n");
}

/**************************************************************************************************
  Timer5 Start Function
**************************************************************************************************/
void Timer5Start(void) {
     //TIM5_CNT = 0;
     CEN_TIM5_CR1_bit = 1;              // Start timer5
     USB_PutQueue("Timer5 started\r\n");
}

/**************************************************************************************************
  Timer5 Stop Function
**************************************************************************************************/
void Timer5Stop(void) {
     CEN_TIM5_CR1_bit = 0;             // Stop timer5
     USB_PutQueue("Timer5 stopped\r\n");
}
//-------------------------------------------------------------------------------------------------


/**************************************************************************************************
  Timer6 Interrupt Handler
  This timer does not have inputs or outputs. It is a 16-bit auto-reload counter driven by a
  16-bit programmable prescaler.
  This timer may be used as a generic timer for time-base generation but it is also specifically
  used to drive the digital-to-analog converter (DAC). In fact, the timer is internally connected
  to the DAC and is able to drive it through its trigger output.
**************************************************************************************************/
void Timer6Interrupt() iv IVT_INT_TIM6_DAC ics ICS_AUTO {
     UIF_TIM6_SR_bit = 0;   // Clear Update Interrupt Flag in TIM6_SR
     //CEN_TIM6_CR1_bit = 0;  // Disable counter in TIM6_CR1
     //USB_PutQueue("Timer6 interrupt\r\n");
     // NOTE: can't use USB_PutQueue from interrupt when using FreeRTOS - hangs up.
     IndTog();
}

/**************************************************************************************************
  Initialize Timer6
  This timer is on APB1 which runs at 60MHz for a CPU frequency of 120MHz.
  Set prescaler to divide by 6000: 60000000 / 6000 = 10000 Hz
  Set auto-reload to divide by 10000: 10000 / 10000 = 1 Hz
  Therefore interrupt will occur once per second
**************************************************************************************************/
void Timer6_Initialize(void) {
     TIM6EN_bit = 1;          // Enable TIM6 clock in RCC_APB1ENR
     CEN_TIM6_CR1_bit = 0;    // Disable timer
     TIM6_PSC = 5999;         // Set the prescaler register to divide by 5999+1=6000
     TIM6_ARR = 10000;        // Set the auto-reload register to divide by 10000
     TIM6_CR2bits.MMS = 0x0;  // Update event is reset
     //OPM_TIM6_CR1_bit = 1;    // One-pulse mode - clears the CEN bit
     NVIC_IntEnable(IVT_INT_TIM6_DAC);  // Set interrupt vector
     UIE_TIM6_DIER_bit = 1;   // Enable TIM6 interrupts in TIM6_DIER
     USB_PutQueue("Timer6 initialized\r\n");
}

/**************************************************************************************************
  Start Timer6 */
void Timer6Start(void) {
     CEN_TIM6_CR1_bit = 1;     // Enable counter in TIM6_CR1
     USB_PutQueue("Timer6 started\r\n");
}

/**************************************************************************************************
  Stop Timer 6
**************************************************************************************************/
void Timer6Stop(void) {
     CEN_TIM6_CR1_bit = 0;    // Disable timer
     USB_PutQueue("Timer6 stopped\r\n");
}
//-------------------------------------------------------------------------------------------------


/**************************************************************************************************
  Timer7 Interrupt Handler
  This timer does not have inputs or outputs. It is a 16-bit auto-reload counter driven by a
  16-bit programmable prescaler.
  This timer may be used as a generic timer for time-base generation but it is also specifically
  used to drive the digital-to-analog converter (DAC). In fact, the timer is internally connected
  to the DAC and is able to drive it through its trigger output.
**************************************************************************************************/
void Timer7Interrupt() iv IVT_INT_TIM7 ics ICS_AUTO {
     UIF_TIM7_SR_bit = 0;   // Clear Update Interrupt Flag in TIM7_SR
     //CEN_TIM7_CR1_bit = 0;  // Disable counter in TIM7_CR1
     //USB_PutQueue("Timer7 interrupt\r\n");
     // NOTE: can't use USB_PutQueue from interrupt when using FreeRTOS - hangs up.
     IndTog();
}

/**************************************************************************************************
  Initialize Timer7
  This timer is on APB1 which runs at 60MHz for a CPU frequency of 120MHz.
  Set prescaler to divide by 6000: 60000000 / 6000 = 10000 Hz
  Set auto-reload to divide by 10000: 10000 / 10000 = 1 Hz
  Therefore interrupt will occur once per second
**************************************************************************************************/
void Timer7_Initialize(void) {
     TIM7EN_bit = 1;          // Enable TIM7 clock in RCC_APB1ENR
     CEN_TIM7_CR1_bit = 0;    // Disable timer
     TIM7_PSC = 5999;         // Set the prescaler register to divide by 5999+1=6000
     TIM7_ARR = 1000;         // Set the auto-reload register to divide by 10000
     TIM7_CR2bits.MMS = 0x0;  // Update event is reset
     //OPM_TIM7_CR1_bit = 1;    // One-pulse mode - clear the CEN bit
     NVIC_IntEnable(IVT_INT_TIM7);  // Set interrupt vector
     UIE_TIM7_DIER_bit = 1;   // Enable TIM7 interrupts in TIM7_DIER

     USB_PutQueue("Timer7 initialized\r\n");
}

/**************************************************************************************************
  Start Timer7
**************************************************************************************************/
void Timer7Start(void) {
     CEN_TIM7_CR1_bit = 1;     // Enable counter in TIM7_CR1
     USB_PutQueue("Timer7 started\r\n");
}

/**************************************************************************************************
  Stop Timer 7
**************************************************************************************************/
void Timer7Stop(void) {
     CEN_TIM7_CR1_bit = 0;    // Disable timer
     USB_PutQueue("Timer7 stopped\r\n");
}
//-------------------------------------------------------------------------------------------------


/**************************************************************************************************
  Advanced-control Timers (TIM8)
  Timer 8 Channel 2 is on APB2 at 120MHz. Channel 2 output is on PC7.
  This timer can be used as a diagnostic aid. Put the Timer8_Ch2_Start() function call anywhere
  in your code to flash PC7 LED.
**************************************************************************************************/
void Timer8_Initialize(void) {
     //
     GPIOCEN_bit = 1;             // Set GPIOC clock enable bit
     GPIOC_MODER &= 0xFFFF3FFF;   // Clear port C MODER-7 bits
     GPIOC_MODER |= 0x00008000;   // Set port C MODER-7 bits to 10: alternate function mode
     GPIOC_AFRL  &= 0x0FFFFFFF;   // Clear port C AFRL-7 bits
     GPIOC_AFRL  |= 0x30000000;   // Set port C AFRL-7 bits to 0011: AF3 TIM8_CH2
     //
     // RCC_APB2ENR
     TIM8EN_bit = 1;               // Enable TIM8 clock in RCC_APB2ENR.B1
     TIM8_CR1  = 0x0008;          // Set OPM bit (one-pulse mode)
     TIM8_PSC  = 119;             // Prescaler value to divide by 120
     TIM8_ARR  = 10000;           // Set auto-reload register to divide by 10000
     TIM8_CCR2 = 20;              // Delay value

     UG_TIM8_EGR_bit = 1;         // Generate update
     //TIM8_EGR = 1;

     TIM8_CCMR1_Output = 0x7000;  // PWM mode 2 for CH2
     TIM8_CCER = 0x0010;          // Channels 2 is connected to output
     TIM8_BDTR = 0x8000;
     USB_PutQueue("Timer8 initialized\r\n");
}

/**************************************************************************************************
  Start Timer 8 Channel 2
**************************************************************************************************/
void Timer8_Ch2_Start(void) {
     CEN_TIM8_CR1_bit = 1;         // Enable timer 8 channel 2
     USB_PutQueue("Timer8 started\r\n");
}
//-------------------------------------------------------------------------------------------------


/**************************************************************************************************
  NOTE: not used in this application
  General-Purpose Timer (TIM10)
  Timer 10 Channel 1 on PB8 is used for PWM control of buzzer.
**************************************************************************************************/
void Timer10_Initialize(void) {

     //TIM10EN_bit = 1;        // Enable clock for timer1

     // Use library to set things up
     RatioBuz = PWM_TIM10_Init(800);   // Set frequency to 800Hz
     //RatioBuz = PWM_TIM10_Init(1000);  // Set frequency to 1000Hz
     PWM_TIM10_Set_Duty(RatioBuz/2, _PWM_NON_INVERTED, _PWM_CHANNEL1);  // Set duty cycle to one-half of period

     USB_PutQueue("Timer10 initialized\r\n");
}

/**************************************************************************************************
  Start Timer 10 - Activate Buzzer
**************************************************************************************************/
void BuzzerOn(void) {
     PWM_TIM10_Start(_PWM_CHANNEL1, &_GPIO_MODULE_TIM10_CH1_PB8);
}

/**************************************************************************************************
  Stop Timer 10 - Deactivate Buzzer
**************************************************************************************************/
void BuzzerOff(void) {
     PWM_TIM10_Stop(_PWM_CHANNEL1);
}

/**************************************************************************************************
  TimeBuzzer - Activate buzzer for a specific time
**************************************************************************************************/
void TimeBuzzer(unsigned Msec) {
     TIM7_ARR = Msec;
     BuzzerOn();
     Timer7Start();
}
//-------------------------------------------------------------------------------------------------


/**************************************************************************************************
  Timer 11 Interrupt Handler
  Using this timer instead of SysTick which is already used by FreeRTOS
**************************************************************************************************/
void Timer11Interrupt() iv IVT_INT_TIM1_TRG_COM_TIM11 ics ICS_AUTO {
     UIF_TIM11_SR_bit = 0;   // Clear Update Interrupt Flag in TIM11_SR

     SysTimeCnt++;

     // For each delay counter if value > 0 subtract 1
     for (CntrPtr = 0; CntrPtr < 2; CntrPtr++) {
          if (DelayCntr[CntrPtr] > 0)
          DelayCntr[CntrPtr]--;
     }

     Msec += 1;                    // Increment millisecond counter
     if (Msec > 999) {             // Count to 1000
          Msec = 0;                 // Set it to 0
          TimerFlag = 1;            // Set 1 second flag
     }
     irqhandler_timer();           // Service timer pool
}

/**************************************************************************************************
  Initialize Timer 11
  Timer11 is on APB2 peripheral clock which runs at 120MHz for a CPU clock of 120MHz.
  Set prescaler to divide by 120: 120000000 / 12000 = 10000 Hz
  Set auto-reload to divide by 10000: 10000 / 10000 = 1 Hz
  Therefore interrupt will occur every second
**************************************************************************************************/
void Timer11_Initialize(void) {
     TIM11EN_bit = 1;              // Enable TIM11 clock in RCC_APB2ENR
     CEN_TIM11_CR1_bit = 0;        // Disable timer while setting registers

     // Set timer to interrupt every millisecond
     TIM11_PSC = 119;              // Set the prescaler register to divide by 119+1=120
     TIM11_ARR = 1000;             // Set the auto-reload register to divide by 1000

     // Set timer to interrupt every second
     //TIM11_PSC = 11999;            // Set the prescaler register to divide by 11999+1=12000
     //TIM11_ARR = 10000;            // Set the auto-reload register to divide by 10000

     UG_TIM11_EGR_bit = 1;         // Generate update

     NVIC_IntEnable(IVT_INT_TIM1_TRG_COM_TIM11);  // Set interrupt vector
     UIE_TIM11_DIER_bit = 1;       // Update interrupt enabled

     USB_PutQueue("Timer11 initialized\r\n");
}

/**************************************************************************************************
  Start Timer 11
**************************************************************************************************/
void Timer11_Start(void) {

     CEN_TIM11_CR1_bit = 1;              // Start timer5
     USB_PutQueue("Timer11 started\r\n");
}

/**************************************************************************************************
  Stop Timer 11
**************************************************************************************************/
void Timer11_Stop(void) {
     CEN_TIM11_CR1_bit = 0;        // Stop timer
     USB_PutQueue("Timer11 stopped\r\n");
}

/**************************************************************************************************
  Generate update on Timer11
**************************************************************************************************/
void Timer11_Update(void) {
     UG_TIM11_EGR_bit = 1;         // Generate update
     USB_PutQueue("Timer11 interrupt generated\r\n");
}
//-------------------------------------------------------------------------------------------------






/**************************************************************************************************
  Diagnostics
**************************************************************************************************/


/**************************************************************************************************
  Display Timer 3 Registers
**************************************************************************************************/
void DisplayTIM3Regs(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"RCC_APB1ENR: %08LX\r\n",RCC_APB1ENR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CR1:    %04X\r\n",TIM3_CR1);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CR2:    %04X\r\n",TIM3_CR2);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_SMCR:   %04X\r\n",TIM3_SMCR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_DIER:   %04X\r\n",TIM3_DIER);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_SR:     %04X\r\n",TIM3_SR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_EGR:    %04X\r\n",TIM3_EGR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CCMR1:  %04X\r\n",TIM3_CCMR1_Output);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CCMR2:  %04X\r\n",TIM3_CCMR2_Output);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CCER:   %04X\r\n",TIM3_CCER);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CNT:    %04X\r\n",TIM3_CNT);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_PSC:    %04X\r\n",TIM3_PSC);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_ARR:    %04X\r\n",TIM3_ARR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CCR1    %04X\r\n",TIM3_CCR1);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CCR2    %04X\r\n",TIM3_CCR2);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CCR3    %04X\r\n",TIM3_CCR3);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM3_CCR4    %04X\r\n",TIM3_CCR4);
     USB_PutQueue(BufNum);

}

/**************************************************************************************************
  Display Timer 4 Registers
**************************************************************************************************/
void DisplayTIM4Regs(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"RCC_APB1ENR: %08LX\r\n",RCC_APB1ENR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CR1:    %04X\r\n",TIM4_CR1);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CR2:    %04X\r\n",TIM4_CR2);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_SMCR:   %04X\r\n",TIM4_SMCR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_DIER:   %04X\r\n",TIM4_DIER);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_SR:     %04X\r\n",TIM4_SR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_EGR:    %04X\r\n",TIM4_EGR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CCMR1:  %04X\r\n",TIM4_CCMR1_Output);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CCMR2:  %04X\r\n",TIM4_CCMR2_Output);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CCER:   %04X\r\n",TIM4_CCER);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CNT:    %04X\r\n",TIM4_CNT);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_PSC:    %04X\r\n",TIM4_PSC);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_ARR:    %04X\r\n",TIM4_ARR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CCR1    %04X\r\n",TIM4_CCR1);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CCR2    %04X\r\n",TIM4_CCR2);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CCR3    %04X\r\n",TIM4_CCR3);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM4_CCR4    %04X\r\n",TIM4_CCR4);
     USB_PutQueue(BufNum);

}

/**************************************************************************************************
  Display Timer 8 Registers
**************************************************************************************************/
void DisplayTIM8Regs(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"RCC_APB2ENR: %08LX\r\n",RCC_APB2ENR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CR1:    %04X\r\n",TIM8_CR1);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CR2:    %04X\r\n",TIM8_CR2);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_SMCR:   %04X\r\n",TIM8_SMCR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_DIER:   %04X\r\n",TIM8_DIER);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_SR:     %04X\r\n",TIM8_SR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_EGR:    %04X\r\n",TIM8_EGR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CCMR1:  %04X\r\n",TIM8_CCMR1_Output);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CCMR2:  %04X\r\n",TIM8_CCMR2_Output);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CCER:   %04X\r\n",TIM8_CCER);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CNT:    %04X\r\n",TIM8_CNT);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_PSC:    %04X\r\n",TIM8_PSC);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_ARR:    %04X\r\n",TIM8_ARR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CCR1    %04X\r\n",TIM8_CCR1);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CCR2    %04X\r\n",TIM8_CCR2);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CCR3    %04X\r\n",TIM8_CCR3);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_CCR4    %04X\r\n",TIM8_CCR4);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"TIM8_BDTR    %04X\r\n",TIM8_BDTR);
     USB_PutQueue(BufNum);
}

/**************************************************************************************************
  Display Timer 10 Registers
**************************************************************************************************/
void DisplayTimer10Regs(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"CR1 register: %04X",TIM10_CR1);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"DIER register: %04X",TIM10_DIER);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"SR register: %04X",TIM10_SR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"EGR register: %04X",TIM10_EGR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"CCMR1 register: %04X",TIM10_CCMR1_Output);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"ARR register: %04X",TIM10_ARR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"CCR1 register: %04X",TIM10_CCR1);
     USB_PutQueue(BufNum);


}

/**************************************************************************************************
  Display GPIOC Registers
**************************************************************************************************/
void DisplayGPIOCRegs(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"GPIOC_MODER    %08LX\r\n",GPIOC_MODER);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"GPIOC_AFRL     %08LX\r\n",GPIOC_AFRL);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"GPIOC_AFRH     %08LX\r\n",GPIOC_AFRH);
     USB_PutQueue(BufNum);
}
//-------------------------------------------------------------------------------------------------


/**************************************************************************************************
  FreeRTOS software timer functions
  These defines have to be included in FreeRTOSConfig.h
#define configUSE_TIMERS                  1
#define configTIMER_TASK_PRIORITY         10
#define configTIMER_QUEUE_LENGTH          5
#define configTIMER_TASK_STACK_DEPTH      5
#define configSUPPORT_DYNAMIC_ALLOCATION  1
**************************************************************************************************/


/**************************************************************************************************
  One-shot timer call-back function
**************************************************************************************************/
void prvOneShotTimerCallback(TimerHandle_t xTimer) {
     TickType_t xTimeNow;
     char *BufNum;

     // Obtain the current tick count.
     xTimeNow = xTaskGetTickCount();

     // Output a string to show the time at which the callback was executed.
     BufNum = GetBuffer();
     sprintf(BufNum,"One-shot time now: %08LX \r\n",xTimeNow);
     USB_PutQueue(BufNum);
}

/**************************************************************************************************
  Create one-shot timer
**************************************************************************************************/
void CreateOneShotTimer(void) {

     // Gives error: Unresolved extern 'xTimerCreate'.
     // The fix was to add 5 more definitions to FreeRTOSConfig.h.
     xOneShotTimer = xTimerCreate("OneShot",500,pdFALSE,0,prvOneShotTimerCallback);
}

/**************************************************************************************************
  Start one-shot timer
**************************************************************************************************/
void StartOneShotTimer(void) {
     BaseType_t xTimerStarted;

     xTimerStarted = xTimerStart(xOneShotTimer, 0);
     if (xTimerStarted == pdPASS) {
          USB_PutQueue("One-shot timer started\r\n");
     } else {
          USB_PutQueue("Error starting one-shot timer\r\n");
     }
}

/**************************************************************************************************
  Change one-shot timer period
**************************************************************************************************/
void ChangeTimerPeriod(TickType_t Period) {
     BaseType_t xTimerChanged;

     xTimerChanged = xTimerChangePeriod(xOneShotTimer, Period, 0);
     if (xTimerChanged == pdPASS) {
          USB_PutQueue("One-shot timer period changed\r\n");
     } else {
          USB_PutQueue("Could not change one-shot timer period\r\n");
     }
}
/*************************************************************************************************/



/**************************************************************************************************
  Set IO output for auto-reload timer (PD8) LED
**************************************************************************************************/
void SetAutoReloadOutput(void) {

     GPIO_Config(&GPIOD_BASE,_GPIO_PINMASK_8, _GPIO_CFG_MODE_OUTPUT | _GPIO_CFG_SPEED_MAX
                 | _GPIO_CFG_OTYPE_PP);
}

/**************************************************************************************************
  Auto-reload timer call-back function
**************************************************************************************************/
void prvAutoReloadTimerCallback(TimerHandle_t xTimer) {

     Ind = ~Ind; // Toggle PD8 LED
}

/**************************************************************************************************
  Create auto-reload timer
**************************************************************************************************/
void CreateAutoReloadTimer(void) {

     SetAutoReloadOutput();
     xAutoReloadTimer = xTimerCreate("AutoReload",500,pdTRUE,0,prvAutoReloadTimerCallback);
}

/**************************************************************************************************
  Start auto-reload timer
**************************************************************************************************/
void StartAutoReloadTimer(void) {
     BaseType_t xTimerStarted;

     xTimerStarted = xTimerStart(xAutoReloadTimer, 0);
     if (xTimerStarted == pdPASS) {
          USB_PutQueue("Auto-reload timer started\r\n");
     } else {
          USB_PutQueue("Error starting auto-reload timer\r\n");
     }
}

/**************************************************************************************************
  Stop auto-reload timer
**************************************************************************************************/
void StopAutoReloadTimer(void) {
     BaseType_t xTimerStopped;

     xTimerStopped = xTimerStop(xAutoReloadTimer, 0);
     if (xTimerStopped == pdPASS) {
          USB_PutQueue("Auto-reload timer stopped\r\n");
     } else {
          USB_PutQueue("Error stopping auto-reload timer\r\n");
     }
}

/**************************************************************************************************
  Change auto-reload timer period
**************************************************************************************************/
void ChangeReloadTimerPeriod(TickType_t Period) {
     BaseType_t xTimer2Changed;

     xTimer2Changed = xTimerChangePeriod(xAutoReloadTimer, Period, 0);
     if (xTimer2Changed == pdPASS) {
          USB_PutQueue("Auto-reload timer period changed\r\n");
     } else {
          USB_PutQueue("Could not change auto-reload timer period\r\n");
     }
}
/*************************************************************************************************/



/**************************************************************************************************
  Diagnostics: PD8 is initialized by the auto-reload timer. If the auto-reload timer is not
  activated, the PD8 output can be used by the following diagnostic functions.
**************************************************************************************************/
void IndOn(void) {
     Ind = 1;       // Set PD8 LED
}

void IndOff(void) {
     Ind = 0;       // Clear PD8 LED
}

void IndTog(void) {
     Ind = ~Ind;    // Toggle PD8 LED
}


/**************************************************************************************************
  Timer task
**************************************************************************************************/
//
void timerTask(void *pvParameters) {

     //StartAutoReloadTimer();
     Timer8_Initialize();  // Initialize diagnostic timer 8 that can flash PC7 LED when started

     while (1) {
          TimerCheck();    // Diagnostic
     }
}
//

/**************************************************************************************************
  End of File
**************************************************************************************************/