/**************************************************************************************************
  SD_Driver.c

  Copyright (C) 2007  Val Gretchev

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************************************/
/*
  MCU:             STM32F407VGT6
  Board:           EasyMx PRO v7 for STM32
  Oscillator:      25 MHz external crystal
  MPU Frequency:   120 MHz from PLL
  Compiler:        mikroC PRO for ARM Version 6.0.0

  NOTE: SPI3 on this board is used to communicate with 4 external peripherals:
    1. Directly with mikroBUS slot 1, and 2
    2. VS1053 audio chip through SW13 bits 1, 2, and 3
    3. SD card socket through SW13 bits 1, 2, and 3
    4. 25P80 flash chip through SW13 bits 1, 2, and 3
**************************************************************************************************/


// Include files
#include <stdint.h>
#include <stddef.h>
#include "FreeRTOSConfig.h"
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"
#include "timers.h"
#include "projdefs.h"
#include "portmacro.h"

#include "SD_Driver.h"
#include "Buffers.h"
#include "RTC.h"
//#include "Common.h"
#include "__Lib_FAT32.h"
#include "MyTimers.h"

#include "USB_Driver3.h"
//#include "UART2_Driver.h"




// Pin definitions for SD card
sbit Mmc_Chip_Select at GPIOD_ODR.B3;  // NOTE: Required by FAT32
sbit SD_CS at GPIOD_ODR.B3;            // Chip Select - PD3 output
sbit SD_CD at GPIOD_IDR.B15;           // Card Detect - PD15 input

// Constants
const code unsigned long CRC32_Tab[] = {
        0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f,
        0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
        0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91, 0x1db71064, 0x6ab020f2,
        0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
        0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9,
        0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
        0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b, 0x35b5a8fa, 0x42b2986c,
        0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
        0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423,
        0xcfba9599, 0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
        0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190, 0x01db7106,
        0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
        0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d,
        0x91646c97, 0xe6635c01, 0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
        0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
        0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
        0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7,
        0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
        0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa,
        0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
        0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81,
        0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
        0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683, 0xe3630b12, 0x94643b84,
        0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
        0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
        0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
        0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5, 0xd6d6a3e8, 0xa1d1937e,
        0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
        0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55,
        0x316e8eef, 0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
        0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe, 0xb2bd0b28,
        0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
        0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f,
        0x72076785, 0x05005713, 0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
        0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
        0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
        0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69,
        0x616bffd3, 0x166ccf45, 0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
        0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc,
        0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
        0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693,
        0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
        0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};



// Variables
char CardOK = FALSE;         // SD card present flag
char FatInit = FALSE;        // Fat32 initialized flag
//char CardWP = FALSE;         // SD write protected flag (not available with micro SD card)
char SDcardStatusChange;     // 0 no change, 1 when a change occurs


__TIME tm;                   // Time structure defined in "__Lib_FAT32.h"
__HANDLE fileHandle[4];      // File handles (4 allowed)
__DIR fsde;                  // Directory structure

// Delete these variables
char FileSM = 0;             // State machine state
char BufferFlag = 0;         // Indicates Buffer is busy when 1


char FileName[16];           // File name
char DisBuffer[128];         // Buffer for displaying file data. NOTE: USB HID uses 64 bytes
char LineBuf[128];           // Buffer for writing a line of data to a file

// Not used
unsigned short int SD_Timer; // Timer handle for this module


short UploadHandle;          // File handle for upload pls.amo sequences
char My_File[8] = "Val.txt"; // Name of file for uploading sequences
char DirectoryName[16];      // Directory name
//char FileName[64];           // File name
char USB_BufCopy[64];
//char USB_Assemble[128];

// To be implemented
QueueHandle_t SDcardQueueHandle;  // Handle for SD card command message queue



unsigned long FileCRC;


extern TTime MyTime;         // Date/time structure in RTC.c
extern unsigned int DelayCntr[];
extern char USB_RdBuff[];
extern char USB_WrBuff[];

extern char TestString[];




/**************************************************************************************************
  External interrupt on PD15 (Card Detect)
**************************************************************************************************/
void ExternalPD15Int() iv IVT_INT_EXTI15_10 ics ICS_AUTO {

     PR15_bit = 1;            // Clear interrupt pending bit
     SDcardStatusChange = 1;  // Set flag when detecting a rising or falling edge
     MR15_bit = 0;            // Mask interrupts
     IndTog();                // Diagnostic
}

/**************************************************************************************************
  External interrupt enable on PD15 (Card Detect)
**************************************************************************************************/
void SD_Card_Detect_Enable(void) {

     NVIC_IntEnable(IVT_INT_EXTI15_10);      // Enable EXTI15 interrupt vector
     // Set SYSCFGEN bit in RCC_APB2ENR register
     SYSCFGEN_bit = 1;                       // Enable system configuration controller clock
     // Set SYSCFG_EXTICR4
     SYSCFG_EXTICR4 |= 0x00003000;           // Connect PD15 to EXTI15
     // Set EXTI_RTSR register
     TR15_bit = 1;                           // Enable rising trigger
     // Set EXTI_FTSR register
     TR15_EXTI_FTSR_bit = 1;                 // Enable falling trigger
     // Clear interrupt
     PR15_bit = 1;                           // Clear interrupt pending bit
     // Set EXTI_IMR register
     MR15_bit = 1;                           // Unmask interrupt request
}

/**************************************************************************************************
  Initialize GPIO
  SD card Chip Select is on PD3  - output
  SD card Card Detect is on PD15 - input
**************************************************************************************************/
void SD_InitIO(void) {

     // SD card control lines
     // Set PD3 (SD_CS) as digital output
     GPIO_Config(&GPIOD_BASE, _GPIO_PINMASK_3, _GPIO_CFG_MODE_OUTPUT | _GPIO_CFG_SPEED_MAX
      | _GPIO_CFG_OTYPE_PP);
     // Set PD15 (SD_CD) as digital input
     GPIO_Config(&GPIOD_BASE, _GPIO_PINMASK_15, _GPIO_CFG_MODE_INPUT | _GPIO_CFG_PULL_UP);

     SD_CS = 1;         // Set inactive level
}

/**************************************************************************************************
  MPU Frequency is 120 MHz from PLL
  SPI3 is on APB1 peripheral clock running at 30MHz
**************************************************************************************************/

/**************************************************************************************************
  Initialize SD card IO and interrupt for Card Detect
**************************************************************************************************/
void SD_Card_Initialize(void) {

     SD_InitIO();
     SD_Card_Detect_Enable();
     CardOK = TRUE;

     //SD_Timer = get_timer();      // Reserve a timer handle for this module
     //USB_PutQueue("SD card IO initialized\r\n");
}

/**************************************************************************************************
  Initialize SPI3 for slow mode.
**************************************************************************************************/
void SPI3_Set_SD_Mode_Slow(void) {

     SPI3_Init_Advanced(_SPI_FPCLK_DIV256, _SPI_MASTER | _SPI_8_BIT | _SPI_CLK_IDLE_LOW
     | _SPI_FIRST_CLK_EDGE_TRANSITION | _SPI_MSB_FIRST | _SPI_SS_DISABLE | _SPI_SSM_ENABLE
     | _SPI_SSI_1, &_GPIO_MODULE_SPI3_PC10_11_12);

}

/**************************************************************************************************
  Initialize SPI3 for fast mode.
**************************************************************************************************/
void SPI3_Set_SD_Mode_Fast(void) {

     //SPI3_Init_Advanced(_SPI_FPCLK_DIV4, _SPI_MASTER | _SPI_8_BIT | _SPI_CLK_IDLE_LOW // Too fast for ADS1246
     SPI3_Init_Advanced(_SPI_FPCLK_DIV16, _SPI_MASTER | _SPI_8_BIT | _SPI_CLK_IDLE_LOW
     //SPI3_Init_Advanced(_SPI_FPCLK_DIV32, _SPI_MASTER | _SPI_8_BIT | _SPI_CLK_IDLE_LOW
     //SPI3_Init_Advanced(_SPI_FPCLK_DIV64, _SPI_MASTER | _SPI_8_BIT | _SPI_CLK_IDLE_LOW
      | _SPI_FIRST_CLK_EDGE_TRANSITION | _SPI_MSB_FIRST | _SPI_SS_DISABLE | _SPI_SSM_ENABLE
      | _SPI_SSI_1, &_GPIO_MODULE_SPI3_PC10_11_12);
}

/**************************************************************************************************
  SPI3 is on APB1 clock operating at 30 MHz.
  Description: Initialize the FAT32.
  Parameters: none
  Returns: 0 - initialized successfully, -1 - error
**************************************************************************************************/
short InitFat32(void) {
     short Status = 0;

     //if (!SD_CD) {
     if (CardOK == TRUE) {

          // Initialize SPI at low speed
          SPI3_Set_SD_Mode_Slow();
          USB_PutQueue("SPI3 set for SD mode slow\r\n");

          Status = FAT32_Init();
          if (Status < 0) {
               USB_PutQueue("FAT32 initialization error\r\n");
               ShowErrorCode();
               FatInit = FALSE;
          } else {
               // Re-initialize SPI at higher speed
               SPI3_Set_SD_Mode_Fast();
               USB_PutQueue("SPI3 set for SD mode fast\r\n");
               FatInit = TRUE;
               USB_PutQueue("FAT32 initialized\r\n");
          }

     } else {
          FatInit = FALSE;
          USB_PutQueue("SD card not available\r\n");
          Status = -1;
     }
     return Status;
}
/*************************************************************************************************/



/**************************************************************************************************
         THESE FUNCTIONS WORK WITH TEXT OR BINARY FILES
**************************************************************************************************/
/**************************************************************************************************
  Description: This function sets the file creation time.
  Parameters: pointer to a time structure
  Returns:  0 - creation time set properly
           -1 - ERROR during function call
**************************************************************************************************/
//short SetGlobalTime(__TIME *pTM) {
short SetGlobalTime(void) {

      RTC_Read(&MyTime);

      tm.Second = MyTime.seconds;
      tm.Minute = MyTime.minutes;
      tm.Hour   = MyTime.hours;
      tm.Day    = MyTime.day;
      tm.Month  = MyTime.month;
      tm.Year   = MyTime.year;

      return (FAT32_SetTime(&tm));
}

/**************************************************************************************************
  Description: This function formats the SD card.
  Parameters: Volume Label (up to 11 characters)
  Returns: 0 - SD card formatted successfully
           1 - SD card unavailable or FAT32 uninitialized
          -1 - error during function call
**************************************************************************************************/
short FormatSD(char *VolumeLabel) {
     char *BufNum;
     short ErrorCode;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          if (FAT32_Format(VolumeLabel) == 0) {
               USB_PutQueue("SD card formatted\r\n");
               return 0;
          } else {
               USB_PutQueue("SD card error during format\r\n");
               ErrorCode = FAT32_GetError();
               BufNum = GetBuffer();
               sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
               USB_PutQueue(BufNum);
               return (-1);
          }
     } else {
          USB_PutQueue("SD card unavailable\r\n");
          return 1;
     }
}

/**************************************************************************************************
  Description: This function changes the directory to the name passed to it.
  Parameters: Directory Name (Parent directory: "..", root directory: "\\".
  Returns: 0 - directory changed successfully
           1 - SD card unavailable or FAT32 uninitialized
          -1 - error during function call
**************************************************************************************************/
short ChangeDir(char *DirName) {
     char *BufNum;
     short ErrorCode;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          if (FAT32_ChangeDir(DirName) == 0) {
               return 0;
          } else {
               USB_PutQueue("SD card error during ChangeDir\r\n");
               ErrorCode = FAT32_GetError();
               BufNum = GetBuffer();
               sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
               USB_PutQueue(BufNum);
               return (-1);
          }
     } else {
          USB_PutQueue("SD card unavailable\r\n");
          return 1;
     }
}

/**************************************************************************************************
  Description: This function makes a directory with the name passed to it.
  Parameters: Directory Name
  Returns: 0 - directory made successfully
           1 - SD card unavailable or FAT32 uninitialized
          -1 - error during function call
**************************************************************************************************/
short MakeDir(char *DirName) {
     char *BufNum;
     short ErrorCode;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          if (FAT32_MakeDir(DirName) == 0) {
               return 0;
          } else {
               USB_PutQueue("SD card error during MakeDir\r\n");
               ErrorCode = FAT32_GetError();
               BufNum = GetBuffer();
               sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
               USB_PutQueue(BufNum);
               return (-1);
          }
     } else {
          USB_PutQueue("SD card unavailable\r\n");
          return 1;
     }
}

/**************************************************************************************************
  Description: This function creates a file for writing with the file name passed to it.
  Parameters: file handle number (0..3), pointer to a file name
  Returns: 0 - file created successfully
           1 - SD card unavailable or FAT32 uninitialized
          -1 - error during function call
**************************************************************************************************/
short CreateFile(unsigned short num, char *filename) {
     char *BufNum;
     short ErrorCode;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          if (FAT32_Exists(filename)) { // Find existing file or create a new one
               fileHandle[num] = FAT32_Open(filename, FILE_APPEND);
               if (fileHandle[num] >= 0) {
                    USB_PutQueue("File opened for append\r\n");
                    return 0;
               } else {
                    USB_PutQueue("File error (append)\r\n");
                    ErrorCode = FAT32_GetError();
                    BufNum = GetBuffer();
                    sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
                    USB_PutQueue(BufNum);
                    return (-1);
               }

          } else {
               fileHandle[num] = FAT32_Open(filename, FILE_WRITE);
               if (fileHandle[num] >= 0) {
                    USB_PutQueue("File opened for write\r\n");
                    return 0;
               } else {
                    USB_PutQueue("File open error (write)\r\n");
                    ErrorCode = FAT32_GetError();
                    BufNum = GetBuffer();
                    sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
                    USB_PutQueue(BufNum);
                    return (-1);
               }
          }

     } else {
          USB_PutQueue("SD card unavailable\r\n");
          return 1;
     }
}

/**************************************************************************************************
  Description: This function closes the file.
  Parameters: file handle number (0..3)
  Returns: 0 - file closed successfully
           1 - SD card unavailable or FAT32 uninitialized
          -1 - error during function call
**************************************************************************************************/
short CloseFile(unsigned short num) {
     char *BufNum;
     short ErrorCode;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          if (FAT32_Close(fileHandle[num]) == 0) {
               USB_PutQueue("File closed\r\n");
               return 0;

          } else {
               USB_PutQueue("Could not close file\r\n");
               ErrorCode = FAT32_GetError();
               BufNum = GetBuffer();
               sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
               USB_PutQueue(BufNum);
               return (-1);
          }
     } else {
          USB_PutQueue("SD card unavailable\r\n");
          return 1;
     }
}

/**************************************************************************************************
  Description: This function opens a file for reading.
  Parameters: file handle number (0..3), pointer to file name
  Returns: 0 - file opened successfully
           1 - SD card unavailable or FAT32 uninitialized
          -1 - file does not exist or error during function call
**************************************************************************************************/
short OpenFile(unsigned short num, char *filename) {
     int8 Status;
     char *BufNum;
     short ErrorCode;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          Status = FAT32_Exists(filename);
          if (Status == 1) {
               fileHandle[num] = FAT32_Open(filename, FILE_READ);
               SD_DEBUGOUT("File opened for read\r\n");
               return 0;
          } else if (Status == 0) {
               USB_PutQueue("File does not exist\r\n");
               return (-1);
          } else if (Status == -1) {
               ErrorCode = FAT32_GetError();
               BufNum = GetBuffer();
               sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
               USB_PutQueue(BufNum);
               return (-1);
          }
     } else {
          USB_PutQueue("SD card unavailable\r\n");
          return 1;
     }
}

/**************************************************************************************************
  Description: Delete a file.
  Parameters: pointer to file name
  Returns: 0 - file deleted successfully
           1 - SD card unavailable or FAT32 uninitialized
          -1 - file does not exist or error during function call
**************************************************************************************************/
short DeleteFile(char *name) {
     int8 Status;
     char *BufNum;
     short ErrorCode;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          //strcpy(FileName,name);

          Status = FAT32_Exists(name);
          if (Status == 1) {
               if (FAT32_Delete(name) == 0) {
                    BufNum = GetBuffer();
                    sprintf(BufNum,"File %s deleted\r\n",name);
                    USB_PutQueue(BufNum);
                    return 0;
               } else {
                    USB_PutQueue("Error FAT32_Delete call\r\n");
                    ErrorCode = FAT32_GetError();
                    BufNum = GetBuffer();
                    sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
                    USB_PutQueue(BufNum);
                    return (-1);
               }
          } else if (Status == 0) {
               BufNum = GetBuffer();
               sprintf(BufNum,"File %s does not exist\r\n",name);
               USB_PutQueue(BufNum);
               return (-1);
          } else if (Status == -1) {
               USB_PutQueue("Error during FAT32_Exists call\r\n");
               ErrorCode = FAT32_GetError();
               BufNum = GetBuffer();
               sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
               USB_PutQueue(BufNum);
               return (-1);
          }

     } else {
          USB_PutQueue("SD card unavailable\r\n");
          return 1;
     }
}

/**************************************************************************************************
  Description: Check if at end of file.
  Parameters: file handle number (0..3)
  Returns: 0 - end of file detected
           1 - SD card unavailable or FAT32 uninitialized
           2 - end of file not detected
          -1 - error during function call
**************************************************************************************************/
short CheckEOF(unsigned short num) {
     int8 Status;
     char *BufNum;
     short ErrorCode;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          Status = (FAT32_Eof(fileHandle[num]));
          if (Status == 1) {
               USB_PutQueue("EOF detected\r\n");
               return 0;
          } else if (Status == 0) {
               USB_PutQueue("Not at EOF\r\n");
               return 2;
          } else if (Status == -1) {
               USB_PutQueue("Error during FAT32_Eof call\r\n");
               ErrorCode = FAT32_GetError();
               BufNum = GetBuffer();
               sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
               USB_PutQueue(BufNum);
               return (-1);
          }

     } else {
          USB_PutQueue("SD card unavailable\r\n");
          return 1;
     }
}
/*************************************************************************************************/




/**************************************************************************************************
         THESE FUNCTIONS WORK ONLY WITH TEXT FILES
**************************************************************************************************/
/**************************************************************************************************
  Description: This function writes a text line to the file from a buffer previously prepared.
  Parameters: file handle number (0..3), pointer to a line buffer set with write data
  Returns: 0 - write was successful
           1 - SD card unavailable or FAT32 uninitialized
          -1 - error during function call
**************************************************************************************************/
short WriteTextLine(unsigned short num, char *buf) {
     int8 Status;
     char *BufNum;
     short ErrorCode;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          Status = FAT32_Write(fileHandle[num], buf, strlen(buf));
          if (Status == 0) {
               USB_PutQueue("Text line written\r\n");
               return 0;
          } else if (Status == -1) {
               USB_PutQueue("Error during FAT32_Write call\r\n");
               ErrorCode = FAT32_GetError();
               BufNum = GetBuffer();
               sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
               USB_PutQueue(BufNum);
               return (-1);
          }

     } else {
          USB_PutQueue("SD card unavailable\r\n");
          return 1;
     }
}

/**************************************************************************************************
  Description: This function reads a line of text data from the opened FILE. Internal pointer is
               advanced to the next line.
  Parameters: file handle number (0..3), pointer to the buffer where data is to be stored
  Returns: 0 - end of file was not reached
           1 - end of file was reached
          -1 - error during function call
**************************************************************************************************/
short ReadTextLine(unsigned short num, char *buf) {
     short Status;
     char *BufNum;
     short ErrorCode;
     int Count;
     char Ch;

     //SPI3_Set_SD_Mode_Fast();
     SPI3_CR1 = SD_SPI_CR1;
     Status = FAT32_Eof(fileHandle[num]);
     if (Status == 1) {
          USB_PutQueue("EOF detected\r\n");
          return 1;
     } else if (Status == 0) {
          memset(buf,0,sizeof(buf));  // Clear the buffer
          Count = 0;
          Ch = 0;

          while ((Ch != '\n') && (Count < 127) && (FAT32_Eof(fileHandle[num]) == 0)) {
               FAT32_Read(fileHandle[num], &Ch, 1);  // Read just one byte
               buf[Count++] = Ch;
          }
          return 0;
     } else if (Status == -1) {
          USB_PutQueue("Error during FAT32_Eof call\r\n");
          ErrorCode = FAT32_GetError();
          BufNum = GetBuffer();
          sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
          USB_PutQueue(BufNum);
          return (-1);
     }
}
/*************************************************************************************************/




/**************************************************************************************************
         THIS FUNCTION WORKS WITH ALL FILES
**************************************************************************************************/
/**************************************************************************************************
  Description: Read the directory and send line by line to USB.
  Parameters: none
  Returns: 0 - write was successful
           1 - SD card unavailable or FAT32 uninitialized
          -1 - error during function call
**************************************************************************************************/
short GetDir(void) {
     int8 Status;
     char *BufNum;
     short ErrorCode;
     char Code = 0;

     if ((CardOK == TRUE) && (FatInit == TRUE)) {

          Status = FAT32_FindFirst(&fsde);
          switch (Status) {

               case FOUND: // 1
                    /* Diagnostic
                    USB_PutQueue("Returned FOUND\r\n");
                    BufNum = GetBuffer();
                    sprintf(BufNum,"Status code: %d\r\n",Status);
                    USB_PutQueue(BufNum);
                    */
                    BufNum = GetBuffer();
                    strcpy(BufNum,fsde.NameExt);  // File name
                    strcat(BufNum,"\r\n");
                    USB_PutQueue(BufNum);
                    break;

               case ERROR: // -1
                    /* Diagnostic
                    USB_PutQueue("Returned ERROR\r\n");
                    BufNum = GetBuffer();
                    sprintf(BufNum,"Status code: %d\r\n",Status);
                    USB_PutQueue(BufNum);
                    */
                    /*
                    ErrorCode = FAT32_GetError();
                    BufNum = GetBuffer();
                    sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
                    USB_PutQueue(BufNum);
                    */
                    ShowErrorCode();
                    break;

               case OK: // 0
                    /* Diagnostic
                    USB_PutQueue("Returned OK\r\n");
                    BufNum = GetBuffer();
                    sprintf(BufNum,"Status code: %d\r\n",Status);
                    USB_PutQueue(BufNum);
                    */
                    USB_PutQueue("No more entries\r\n");
                    break;
          }

          //do {
          while (Status == FOUND) {

               Status = FAT32_FindNext(&fsde);
               switch (Status) {

                    case FOUND: // 1
                         /* Diagnostic
                         USB_PutQueue("Returned FOUND step 2\r\n");
                         BufNum = GetBuffer();
                         sprintf(BufNum,"Status code: %d\r\n",Status);
                         USB_PutQueue(BufNum);
                         */
                         BufNum = GetBuffer();
                         strcpy(BufNum,fsde.NameExt);  // File name
                         strcat(BufNum,"\r\n");
                         USB_PutQueue(BufNum);
                         break;

                    case ERROR: // -1
                         /*
                         USB_PutQueue("Returned ERROR step 2\r\n");
                         BufNum = GetBuffer();
                         sprintf(BufNum,"Status code: %d\r\n",Status);
                         USB_PutQueue(BufNum);

                         ErrorCode = FAT32_GetError();
                         BufNum = GetBuffer();
                         sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
                         USB_PutQueue(BufNum);
                         */
                         ShowErrorCode();
                         break;

                    case OK: // 0
                         /* Diagnostic
                         USB_PutQueue("Returned OK step 2\r\n");
                         BufNum = GetBuffer();
                         sprintf(BufNum,"Status code: %d\r\n",Status);
                         USB_PutQueue(BufNum);
                         */
                         USB_PutQueue("No more entries\r\n");
                         break;
               }
           }
          //} while (Status == FOUND);



     } else {
          if (CardOK == FALSE) {
               USB_PutQueue("SD card not available\r\n");
               Code = 1;
          } else {
          if (FileSM == FALSE)
               USB_PutQueue("FAT32 not initialized\r\n");
               Code = 1;
          }
     }
     return Code;
}
/*************************************************************************************************/














/**************************************************************************************************
         THESE FUNCTIONS WORK ONLY WITH TEXT FILES
**************************************************************************************************/
/**************************************************************************************************
  Description: Read a text file and send to USB line by line.
  Parameters: pointer to file name
  Returns: 0 - success
          -1 - SD card unavailable or error encountered
**************************************************************************************************/
char GetFile(char *name) {
     static short fhandle;
     short Status;
     int Count;
     char Ch;
     char *BufNum;

     //SD_DEBUGOUT(name);
     // Do not proceed if there is no card inserted
     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          SD_DEBUGOUT("SD card is available\r\n");

          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          strcpy(FileName,name);

          fhandle = FAT32_Open(FileName, FILE_READ);
          if (fhandle < 0) {
               SD_DEBUGOUT("Error opening file\r\n");
               return (-1);
          } else {
               SD_DEBUGOUT("File opened\r\n");
               Status = FAT32_Eof(fhandle);
               if (Status == 1) {
                    SD_DEBUGOUT("EOF detected - file empty\r\n");
                    return 0;
               } else if (Status == 0) {
                    while (FAT32_Eof(fhandle) == 0) {
                         memset(DisBuffer,0,sizeof(DisBuffer));  // Clear the buffer
                         Count = 0;
                         Ch = 0;
                         while ((Ch != '\n') && (Count < 128) && (FAT32_Eof(fhandle) == 0)) {
                              FAT32_Read(fhandle, &Ch, 1);
                              DisBuffer[Count++] = Ch;  // Put char into buffer
                         }
                         BufNum = GetBuffer();    // Get a free buffer
                         strcpy(BufNum,DisBuffer);   // Copy to free buffer
                         //USB_PutQueue(BufNum);    // Send free buffer on USB
                         UART2_PutQueue(BufNum);
                    }
                    SD_DEBUGOUT("EOF reached\r\n");
                    FAT32_Close(fhandle);
                    SD_DEBUGOUT("File closed\r\n");
                    return 0;

               } else if (Status == -1) {
                    SD_DEBUGOUT("Error on FAT32_Eof\r\n");
                    return (-1);
               }
          }

     } else {
          USB_PutQueue("SD card not available\r\n");
          return (-1);
     }
}

/**************************************************************************************************
  Description: Opens a file or creates a new file, writes a single text line of data to a file,
               then closes the file.
  Parameters: pointer to file name
  Returns: 0 - line written to file successfully
          -1 - SD card unavailable or error encountered
**************************************************************************************************/
char WriteLine(char *name) {
     static short fhandle;
     char *BufNum;

     // Do not proceed if there is no card inserted
     if ((CardOK == TRUE) && (FatInit == TRUE)) {
          SD_DEBUGOUT("SD card is available\r\n");

          //SPI3_Set_SD_Mode_Fast();
          SPI3_CR1 = SD_SPI_CR1;
          strcpy(FileName,name);

          if (FAT32_Exists(FileName)) {     // Find existing file or create a new one
               fhandle = FAT32_Open(FileName, FILE_APPEND);
               if (fhandle < 0) {
                    BufNum = GetBuffer();
                    sprintf(BufNum,"File append error: %d\r\n",fhandle);
                    USB_PutQueue(BufNum);
                    return (-1);

                 } else {
                     USB_PutQueue("File opened for append\r\n");
                 }
          } else {
               fhandle = FAT32_Open(FileName, FILE_WRITE);
               if (fhandle < 0) {
                    BufNum = GetBuffer();
                    sprintf(BufNum,"File open error: %d\r\n",fhandle);
                    USB_PutQueue(BufNum);
                    return (-1);
               } else {
                    SetGlobalTime(); // Set the file creation date-time
                    SD_DEBUGOUT("File created\r\n");
               }
          }
          FAT32_Write(fhandle, LineBuf, strlen(LineBuf));
          SD_DEBUGOUT("File line written\r\n");
          memset(LineBuf, 0, sizeof(LineBuf));
          FAT32_Close(fhandle);
          SD_DEBUGOUT("File closed\r\n");
          return 0;

     } else {
          return (-1);
     }
}
/*************************************************************************************************/








// This state machine is not used.
//--------------------------------

void DoFileSM(void) {
     int Count;
     char Ch;
     char *BufNum;
     short Status;
     char BinCount;
     static short fhandle;

     switch (FileSM) {

//------ Open a binary file, output the file line by line to USB then close the file  -------------
         case 13:
             SPI3_CR1 = SD_SPI_CR1;
             fhandle = FAT32_Open(FileName, FILE_READ);
             if (fhandle < 0) {
                 USB_PutQueue("Error opening file\r\n");
                 FileSM = 0;
             } else {
                 USB_PutQueue("File opened\r\n");
                 FileSM = 14;
             }
             break;
         case 14: // Check if at end of file
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_Eof(fhandle) > 0) {
                 USB_PutQueue("EOF detected\r\n");
                 FileSM = 17; // EOF detected
             } else {
                 //USB_PutQueue("More data available\r\n");
                 FileSM = 15; // More data available
             }
             break;
         case 15: // Get a line of data and send it on USB
             SPI3_CR1 = SD_SPI_CR1;
             memset(DisBuffer,0,sizeof(DisBuffer)); // Clear the buffers
             memset(LineBuf,0,sizeof(LineBuf));
             FAT32_Read(fhandle,&BinCount,1); // Get the byte count for the line
             if (BinCount > 0) {
                 LineBuf[0] = BinCount; // Put byte count in buffer
                 FAT32_Read(fhandle,LineBuf+1,BinCount); // Get the rest of the bytes
                 //BufNum = GetBuffer();
                 //BinToAscii(BinData,BufNum,BinCount+1);
                 //strcat(BufNum,"\r\n");
                 //USB_PutQueue(BufNum); // Send Buffer on USB
                 BinToAscii(LineBuf,DisBuffer,BinCount+1);
                 strcat(DisBuffer,"\r\n");
                 BufferFlag = 1;
                 USB_PutQueue(DisBuffer);
             } else {
                 USB_PutQueue("Encountered a line of 0 length\r\n");
             }
             FileSM = 16;
             break;
         case 16: // Wait for USB to finish sending the line
             //if (DisBuffer[0] == 0) {
             if (BufferFlag == 0) {
                 FileSM = 14;
             }
             break;
         case 17: // EOF detected - close the file
             SPI3_CR1 = SD_SPI_CR1;
             FAT32_Close(fhandle);
             USB_PutQueue("File closed\r\n");
             FileSM = 0;
             break;


//------ Open a file, write a single line of binary data then close the file ----------------------
//------ Expecting FileName to hold the name of the file being written to -------------------------
//------ Expecting LineBuf to hold the prepared binary data line ----------------------------------
         case 18: // Open requested file for writing
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_Exists(FileName)) { // Find existing file or create a new one
                 fhandle = FAT32_Open(FileName, FILE_APPEND);
                 if (fhandle < 0) {
                     BufNum = GetBuffer();
                     sprintf(BufNum,"File append error: %d\r\n",fhandle);
                     USB_PutQueue(BufNum);
                     FileSM = 0;
                     break;
                 } else {
                     USB_PutQueue("File opened for append\r\n");
                     FileSM = 19;
                 }
             } else {
                 fhandle = FAT32_Open(FileName, FILE_WRITE);
                 if (fhandle < 0) {
                     BufNum = GetBuffer();
                     sprintf(BufNum,"File open error: %d\r\n",fhandle);
                     USB_PutQueue(BufNum);
                     FileSM = 0;
                     break;
                 } else {
                     SetGlobalTime(); // Set the file creation date-time
                     USB_PutQueue("File created\r\n");
                     FileSM = 19;
                 }
             }
             break;
         case 19: // Write the prepared line to the file
             SPI3_CR1 = SD_SPI_CR1;
             FAT32_Write(fhandle,LineBuf,LineBuf[0]+1);
             USB_PutQueue("File line written\r\n");
             memset(LineBuf,0,sizeof(LineBuf));
             FileSM = 20;
             break;
         case 20: // Close the file
             SPI3_CR1 = SD_SPI_CR1;
             FAT32_Close(fhandle);
             USB_PutQueue("File closed\r\n");
             FileSM = 0;
             break;


         //****************************************************************************************









/**************************************************************************************************
         These states were added for uploading files to SD card over USB interface.
         The states are entered from the USB command processor when control characters are sent
         by the host PC.
**************************************************************************************************/

         // Send abort message (NAK) **************************************************************
         case 80:
             USB_PutQueue("Sending NAK\r\n");
             BufNum = GetBuffer();
             BufNum[0] = NAK; // Respond with Negative Acknowledge
             BufNum[1] = 0;
             //USB_PutQueue(BufNum); // Disabled
             FileSM = 0;
             break;
         //****************************************************************************************


         // Holding state while file is open ******************************************************
         case 90: // Set time for timeout value
             set_timer(SD_Timer,1000);
             FileSM = 91;
             break;
         case 91:
             if (check_timer(SD_Timer) == 0) {
                 USB_PutQueue("Hung timeout\r\n");
                 FileSM = 92; // Must be hung up - go close the file
             }
             break;
         case 92: // Close the file handle
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_Close(UploadHandle) < 0) {
                 USB_PutQueue("Error closing pls.amo file\r\n");
                 ShowErrorCode();
                 FileSM = 0;
             } else {
                 USB_PutQueue("File closed after timeout\r\n");
                 USB_PutQueue("Sending abort message\r\n");
                 FileSM = 80; // Send NAK
             }
             break;
         //****************************************************************************************



         // Change directory to the name given in message *****************************************
         case 100: // First make sure we are in the root directory
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_ChangeDir("\\") < 0) {
                 USB_PutQueue("Error changing to root directory\r\n");
                 ShowErrorCode();
                 FileSM = 80; // Send NAK
             } else {
                 USB_PutQueue("Directory changed to root\r\n");
                 FileSM = 101;
             }
             break;
         case 101: // Check if directory exists or not
             SPI3_CR1 = SD_SPI_CR1;
             Status = FAT32_Exists(DirectoryName);
             switch (Status) {
                 case 0:
                     BufNum = GetBuffer();
                     sprintf("Directory %s does not exist on SD card\r\n",DirectoryName);
                     USB_PutQueue(BufNum);
                     FileSM = 102; // Make a new directory
                     break;
                 case 1:
                     BufNum = GetBuffer();
                     sprintf("Directory %s already exists on SD card\r\n",DirectoryName);
                     USB_PutQueue(BufNum);
                     FileSM = 103; // Change directory
                     break;
                 case (-1):
                     USB_PutQueue("Error in FAT32_Exists function call\r\n");
                     ShowErrorCode();
                     FileSM = 80; // Send NAK
                     break;
             }
             break;
         case 102: // Directory does not exist - make a new one
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_MakeDir(DirectoryName) < 0) {
                 BufNum = GetBuffer();
                 sprintf("Error making directory %s\r\n",DirectoryName);
                 USB_PutQueue(BufNum);
                 ShowErrorCode();
                 FileSM = 80; // Send NAK
             } else {
                 BufNum = GetBuffer();
                 sprintf("Directory %s created\r\n",DirectoryName);
                 USB_PutQueue(BufNum);
                 FileSM = 103;
             }
             break;
         case 103: // Change directory to the name supplied
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_ChangeDir(DirectoryName) < 0) {
                 USB_PutQueue("Error changing directory\r\n");
                 ShowErrorCode();
                 FileSM = 80; // Send NAK
             } else { // Send DC1
                 BufNum = GetBuffer();
                 sprintf(BufNum,"Directory changed to %s\r\n",DirectoryName);
                 USB_PutQueue(BufNum);
                 FileSM = 104;
             }
             break;
         case 104: // Erase file in current directory
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_Exists(My_File)) {
                 if (FAT32_Delete(My_File) < 0) {
                     USB_PutQueue("Error deleting file\r\n");
                     ShowErrorCode();
                     FileSM = 80; // Send NAK
                 } else {
                     USB_PutQueue("File deleted\r\n");
                     FileSM = 105;
                 }
             } else {
                 USB_PutQueue("File does not exist on SD card\r\n");
                 FileSM = 105;
             }
             break;
         case 105:
             BufNum = GetBuffer();
             BufNum[0] = DC1; // Respond with Device Control 1
             BufNum[1] = 0;
             //USB_PutQueue(BufNum); // Disabled
             FileSM = 0;
             break;
         //****************************************************************************************


         // Open a file for writing ***************************************************************
         case 110: // Open file and send back STX response
             SPI3_CR1 = SD_SPI_CR1;
             UploadHandle = FAT32_Open(My_File,FILE_WRITE);
             if (UploadHandle < 0) {
                 USB_PutQueue("Error opening file\r\n");
                 ShowErrorCode();
                 FileSM = 80; // Send NAK
             } else {
                 FileCRC = 0xFFFFFFFF; // Set initial CRC seed value
                 //set_timer(SD_Timer,USB_DELAY);
                 //FileSM = 111;
                 FileSM = 112; // Bypass delay
             }
             break;
         case 111: // Delay
             if (check_timer(SD_Timer) == 0) {
                 FileSM = 112;
             }
             break;
         case 112:
             USB_PutQueue("File opened for writing\r\n");
             BufNum = GetBuffer();
             BufNum[0] = STX; // Respond with Start Of Text
             BufNum[1] = 0;
             //USB_PutQueue(BufNum); // Disabled
             FileSM = 90; // Wait with open file
             break;
         //****************************************************************************************


         // Close a file upon command from host ***************************************************
         case 120: // Close the file handle
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_Close(UploadHandle) < 0) {
                 USB_PutQueue("Error closing file\r\n");
                 ShowErrorCode();
                 FileSM = 80; // Send NAK
             } else {
                 USB_PutQueue("File closed\r\n");
                 BufNum = GetBuffer();
                 BufNum[0] = ETX; // Respond with End Of Text
                 BufNum[1] = 0;
                 //USB_PutQueue(BufNum); // Disabled
                 FileSM = 121;
             }
             break;
         case 121: // First make sure we are in the root directory
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_ChangeDir("\\") < 0) {
                 USB_PutQueue("Error changing to root directory\r\n");
                 ShowErrorCode();
                 FileSM = 80; // Send NAK
             } else {
                 USB_PutQueue("Directory changed to root\r\n");
                 FileSM = 122;
             }
             break;
         case 122: // Display CRC
             FileCRC ^= 0xFFFFFFFF; // Final exclusive or
             BufNum = GetBuffer();
             sprintf(BufNum,"FileCRC: 0x%08LX\r\n",FileCRC);
             USB_PutQueue(BufNum);
             FileSM = 0;
             break;
         //****************************************************************************************


         // Process first part of long line *******************************************************
         case 130: // Save 1st partial line in LeneBuf
             FileCRC = CRC32Calc(FileCRC,USB_BufCopy,strlen(USB_BufCopy));

             strcpy(LineBuf,USB_BufCopy);
             //set_timer(SD_Timer,USB_DELAY);
             //FileSM = 131;
             FileSM = 132; // Bypass delay
             break;
         case 131: // Delay
             if (check_timer(SD_Timer) == 0) {
                 FileSM = 132;
             }
             break;
         case 132: // Respond to application
             BufNum = GetBuffer();
             BufNum[0] = DC2; // Respond with Device Control 2
             BufNum[1] = 0;
             //USB_PutQueue(BufNum); // Disabled
             FileSM = 90; // Wait with file open
             break;
         //****************************************************************************************


         // Process second part of long line ******************************************************
         case 140: // Add 2nd partial line to LineBuf and write to SD card
             FileCRC = CRC32Calc(FileCRC,USB_BufCopy,strlen(USB_BufCopy));

             SPI3_CR1 = SD_SPI_CR1;
             strcat(LineBuf,USB_BufCopy);
             StripCtrlCharExceptTab(LineBuf);
             strcat(LineBuf,"\r\n"); // Add CRLF
             if (FAT32_Write(UploadHandle,LineBuf,strlen(LineBuf)) < 0) {
                 USB_PutQueue("Error writing to file\r\n");
                 ShowErrorCode();
                 FileSM = 160; // Close the file
             } else {
                 //set_timer(SD_Timer,USB_DELAY);
                 //FileSM = 141;
                 FileSM = 142; // Bypass delay
             }
             break;
         case 141: // Delay
             if (check_timer(SD_Timer) == 0) {
                 FileSM = 142;
             }
             break;
         case 142: // Respond to application
             BufNum = GetBuffer();
             BufNum[0] = DC3; // Respond with Device Control 3
             BufNum[1] = 0;
             //USB_PutQueue(BufNum); // Disabled
             FileSM = 90; // Wait with file open
             break;
         //****************************************************************************************


         // Process short line ********************************************************************
         case 150: // Move whole line to LineBuf and write to SD card
             FileCRC = CRC32Calc(FileCRC,USB_BufCopy,strlen(USB_BufCopy));

             SPI3_CR1 = SD_SPI_CR1;
             strcpy(LineBuf,USB_BufCopy);
             StripCtrlCharExceptTab(LineBuf);
             strcat(LineBuf,"\r\n"); // Add CRLF
             if (FAT32_Write(UploadHandle,LineBuf,strlen(LineBuf)) < 0) {
                 USB_PutQueue("Error writing to file\r\n");
                 ShowErrorCode();
                 FileSM = 160; // Close the file
             } else {
                 //set_timer(SD_Timer,USB_DELAY);
                 //FileSM = 151;
                 FileSM = 152; // Bypass delay
             }
             break;
         case 151:
             if (check_timer(SD_Timer) == 0) {
                 FileSM = 152;
             }
             break;
         case 152:// Respond to application
             BufNum = GetBuffer();
             BufNum[0] = DC4;; // Respond with Device Control 4
             BufNum[1] = 0;
             //USB_PutQueue(BufNum); // Disabled
             FileSM = 90; // Wait with file open
             break;
         //****************************************************************************************


         // Close file in case of error during write **********************************************
         case 160: // Close the file handle
             SPI3_CR1 = SD_SPI_CR1;
             if (FAT32_Close(UploadHandle) < 0) {
                 USB_PutQueue("Error closing file\r\n");
                 ShowErrorCode();
                 FileSM = 80; // Send NAK
             } else {
                 USB_PutQueue("File closed\r\n");
                 FileSM = 80; // Send NAK
             }
             break;
         //****************************************************************************************





     }
}


/**************************************************************************************************
         CONVERSION FUNCTIONS FOR TRANSLATING BINARY LINE TO ASCII AND ASCII LINE TO BINARY
**************************************************************************************************/
/**************************************************************************************************
  Convert a line of binary data into ASCII.
  InputBuf points to binary data in hexadecimal format.
  OutputBuf points to an ASCII data buffer.
  ByteCount is the length of the input buffer.
  Returns 0 if ByteCount is greater than 0.
  Returns -1 if ByteCount is 0.
  OutputBuf size will always be twice as long as InputBuf.
**************************************************************************************************/
short BinToAscii(char *InputBuf, char *OutputBuf, char ByteCount) {
     int i; // Counter

     if (ByteCount > 0) {
          for (i=0;i<ByteCount;i++) {
               ByteToHex(InputBuf[i],&OutputBuf[i*2]);  // Convert binary byte into 2 ASCII
          }                                // characters and enter them into output buffer
          return (0);
     } else {
          return (-1);
     }
}

/**************************************************************************************************
  Format a line of ASCII into binary data.
  InputBuf points to ASCII data that contains characters 0 through 9 and A through F.
  OutputBuf points to a binary data buffer.
  ByteCount is the length of the input buffer divided by 2.
  Returns 0 if ByteCount is greater than 0.
  Returns -1 if ByteCount is 0.
  OutputBuf size will always be half as long as InputBuf.
**************************************************************************************************/
short AsciiToBin(char *InputBuf, char *OutputBuf, char ByteCount) {
     int i;           // Counter
     char AscStr[4];  // Holds two-byte string from InputBuf for conversion by xtoi

     if (ByteCount > 0) {
          for (i=0;i<ByteCount;i++) {        // Half the count of the ASCII buffer
               AscStr[0] = InputBuf[i*2];    // Get 2 bytes at a time from input buffer
               AscStr[1] = InputBuf[i*2+1];
               AscStr[2] = 0;
               OutputBuf[i] = xtoi(AscStr);  // Convert 2 ASCII HEX characters to binary byte
          }                                  // and enter it into output buffer
          return (0);
     } else {
          return (-1);
     }
}
/*************************************************************************************************/


/**************************************************************************************************
         THESE FUNCTIONS WORK ONLY WITH BINARY FILES
**************************************************************************************************/
/**************************************************************************************************
  Description: This function writes a line of binary data to the opened FILE. Internal pointer is
               advanced to the next line. Byte 0 of buf is assumed to contain the byte count
               of the data bytes not including the byte count itself.
  Parameters: file handle number (0..3), pointer to the buffer where data is located.
  Prerequisites: File must be opened for writing or appending.
  Returns: 0 - write was successful
          -1 - error during function call
**************************************************************************************************/
short WriteBinLine(unsigned short num, char *buf) {
     //SPI3_Set_SD_Mode_Fast();
     SPI3_CR1 = SD_SPI_CR1;
     if (FAT32_Write(fileHandle[num],buf,buf[0]+1) == 0)
          return 0;
     else
          return (-1);
}

/**************************************************************************************************
  Description: This function reads a line of binary data from the opened file. Internal pointer is
               advanced to the next line. Byte 0 of buf is assumed to contain the byte count
               of the data bytes not including the byte count itself.
  Parameters: file handle number (0..3), pointer to the buffer where data is to be stored.
  Prerequisites: File must be opened for reading.
  Returns: 0 - end of file was not reached
           1 - end of file was reached
          -1 - error during function call
**************************************************************************************************/
short ReadBinLine(unsigned short num, char *buf) {
     short Status;
     char Count;

     //SPI3_Set_SD_Mode_Fast();
     SPI3_CR1 = SD_SPI_CR1;
     Status = FAT32_Eof(fileHandle[num]);
     if (Status == 0) {
          memset(buf,0,sizeof(buf));                    // Clear binary data Buffer
          FAT32_Read(fileHandle[num],&Count,1);         // Get the byte count for the line
          if (Count > 0) {
               buf[0] = Count;                          // Put byte count in buffer
               FAT32_Read(fileHandle[num],buf+1,Count); // Get rest of bytes and put then in buffer
          } else if (Status == 1) {
               USB_PutQueue("End of file\r\n");
          } else {
               USB_PutQueue("Error reading file\r\n");
          }
     }
     return (Status);
}
/**************************************************************************************************
  Read a binary file of name and send it out on USB.
**************************************************************************************************/
void ReadWholeBinFile(char *name) {
     char *BufNum;

     memset(LineBuf,0,sizeof(LineBuf)); // Clear binary data buffer
     strcpy(FileName,name);

     //SPI3_Set_SD_Mode_Fast();
     SPI3_CR1 = SD_SPI_CR1;
     fileHandle[2] = FAT32_Open(FileName, FILE_READ);
     if (fileHandle[2] >= 0) {
          while (FAT32_Eof(fileHandle[2]) == 0) {
               BufNum = GetBuffer();
               FAT32_Read(fileHandle[2],LineBuf,6);
               sprintf(BufNum,"%d%d%d%d%d%d\r\n",LineBuf[0],LineBuf[1],LineBuf[2],LineBuf[3],
                                                 LineBuf[4],LineBuf[5]);
               USB_PutQueue(BufNum);
          }
          FAT32_Close(fileHandle[2]);
     } else {
          USB_PutQueue("Error opening file\r\n");
     }
}



/**************************************************************************************************
  Test binary to ASCII conversion. Call this function in USB.c
  Should display 000155AA on USB.
**************************************************************************************************/
void TestBinToAsc(void) {

     memset(DisBuffer,0,sizeof(DisBuffer));  // Clear ASCII data buffer

     LineBuf[0] = 0x00;                // Enter some binary data into line buffer
     LineBuf[1] = 0x01;
     LineBuf[2] = 0x55;
     LineBuf[3] = 0xAA;
     BinToAscii(LineBuf,DisBuffer,4);     // Convert binary data into ASCII
     strcat(DisBuffer,"\r\n");            // Append line ending required by USB
     USB_PutQueue(DisBuffer);           // Display the ASCII buffer
}

/**************************************************************************************************
  Test ASCII to binary conversion. Call this function in USB.c
  Should display AA55BB12 on USB.
**************************************************************************************************/
void TestAscToBin(void) {
     char *BufNum;

     memset(LineBuf,0,sizeof(LineBuf)); // Clear binary data buffer

     strcpy(DisBuffer,"AA55BB12");        // Put an ASCII string into buffer
     AsciiToBin(DisBuffer,LineBuf,4);     // Convert ASCII to binary

     BufNum = GetBuffer();             // Have to convert it back to ASCII to display
     sprintf(BufNum,"%02X%02X%02X%02X\r\n",LineBuf[0],LineBuf[1],LineBuf[2],LineBuf[3]);
     USB_PutQueue(BufNum);

}


/**************************************************************************************************
  Testing write binary file.
**************************************************************************************************/
void WriteBinFile(void) {

     memset(LineBuf,0,sizeof(LineBuf)); // Clear binary data buffer
     strcpy(FileName,"TRY1.BIN");

     LineBuf[0] = 5; // Count
     LineBuf[1] = 0; // 5 bytes of data
     LineBuf[2] = 1;
     LineBuf[3] = 2;
     LineBuf[4] = 3;
     LineBuf[5] = 4;

     //SPI3_Set_SD_Mode_Fast();
     SPI3_CR1 = SD_SPI_CR1;
     if (FAT32_Exists(FileName)) {
          fileHandle[2] = FAT32_Open(FileName, FILE_APPEND);
     } else {
          fileHandle[2] = FAT32_Open(FileName, FILE_WRITE);
     }
     if (fileHandle[2] >= 0) {
          FAT32_Write(fileHandle[2],LineBuf,6);
          FAT32_Close(fileHandle[2]);
          USB_PutQueue("Line written\r\n");
     } else {
          USB_PutQueue("Error opening file\r\n");
     }
}

/**************************************************************************************************
  Testing read binary file.
**************************************************************************************************/
void ReadBinFile(void) {
     char *BufNum;

     memset(LineBuf,0,sizeof(LineBuf)); // Clear binary data buffer
     strcpy(FileName,"TRY1.BIN");

     //SPI3_Set_SD_Mode_Fast();
     SPI3_CR1 = SD_SPI_CR1;
     fileHandle[2] = FAT32_Open(FileName, FILE_READ);
     if (fileHandle[2] >= 0) {
          while (FAT32_Eof(fileHandle[2]) == 0) {
               BufNum = GetBuffer();
               FAT32_Read(fileHandle[2],LineBuf,6);
               sprintf(BufNum,"%d%d%d%d%d%d\r\n",LineBuf[0],LineBuf[1],LineBuf[2],LineBuf[3],
                                                 LineBuf[4],LineBuf[5]);
               USB_PutQueue(BufNum);
          }
          FAT32_Close(fileHandle[2]);
     } else {
          USB_PutQueue("Error opening file\r\n");
     }
}

/**************************************************************************************************
  Diagnostics
**************************************************************************************************/
// Display SPI3 registers. Call this function from USB command processor.
void ShowSPI3Regs(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"SPI3_CR1 register:     %04X\r\n",SPI3_CR1);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"SPI3_CR2 register:     %04X\r\n",SPI3_CR2);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"SPI3_SR register:      %04X\r\n",SPI3_SR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"SPI3_CRCPR register:   %04X\r\n",SPI3_CRCPR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"SPI3_RXCRCR register:  %04X\r\n",SPI3_RXCRCR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"SPI3_TXCRCR register:  %04X\r\n",SPI3_TXCRCR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"SPI3_I2SCFGR register: %04X\r\n",SPI3_I2SCFGR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"SPI3_I2SPR register:   %04X\r\n",SPI3_I2SPR);
     USB_PutQueue(BufNum);
}

// Check for presence of SD card
void CheckStateOfCardDetect(void) {

     if (SD_CD)
          USB_PutQueue("SD card is withdrawn\r\n");
     else
          USB_PutQueue("SD card is inserted\r\n");
}

/**************************************************************************************************
  Strip control character from string.
**************************************************************************************************/
short StripControlChar(char *StrIn, char *StrOut, int Count) {
     char i;

     memset(StrOut,0,sizeof(StrOut));
     for (i=0;i<Count;i++) {
          StrOut[i] = StrIn[i+1]; // Just remove control character at the start of line
     }
     return i;
}

/**************************************************************************************************
  Strip control character from string but leave TAB character in data.
**************************************************************************************************/
void StripCtrlCharExceptTab(char* StrIn) {
     char CleanBuf[256] = {0};
     char i,j;
     char Ch;

     for (i=0,j=0;i<sizeof(StrIn);i++) {
          Ch = StrIn[i];
          if ((Ch == '\t') || (Ch > 31)) {
              CleanBuf[j++] = Ch;
          }
     }
     for (i=0;i<sizeof(StrIn);i++) {
          StrIn[i] = CleanBuf[i];
     }
}

/**************************************************************************************************
  CRC32 calculation.
**************************************************************************************************/

unsigned long CRC32Calc(unsigned long vCRC32, char *Buff, int Size) {
     char *BufNum;
     unsigned long CRC;
     int i;

     CRC = vCRC32;
     //
     BufNum = GetBuffer();
     sprintf(BufNum,"Initial CRC: 0x%08LX\r\n",CRC);
     USB_PutQueue(BufNum);
     //
     //
     BufNum = GetBuffer();
     sprintf(BufNum,"Length of HW: %d\r\n",strlen(Buff));
     USB_PutQueue(BufNum);
     //

     for (i=0;i<strlen(Buff);i++) {
          CRC = CRC32_Tab[(CRC ^ Buff[i]) & 0xFF] ^ (CRC >> 8);
     }

     //
     BufNum = GetBuffer();
     sprintf(BufNum,"Final CRC:   0x%08LX\r\n",CRC);
     USB_PutQueue(BufNum);
     //
     return CRC;
}

/**************************************************************************************************
  Display error code on USB.
**************************************************************************************************/
void ShowErrorCode(void) {
     char *BufNum;
     short ErrorCode;

     SPI3_CR1 = SD_SPI_CR1;
     ErrorCode = FAT32_GetError();
     if (ErrorCode < 0) {
          BufNum = GetBuffer();
          sprintf(BufNum,"Error code: %d\r\n",ErrorCode);
          USB_PutQueue(BufNum);
     } else {
          USB_PutQueue("Error code is 0 - no error\r\n");
     }
}

/**************************************************************************************************
  SD card detect switch debounce subroutine
**************************************************************************************************/
void SDcardDetect(void) {
     if (SDcardStatusChange == 1) {
          SDcardStatusChange = 0;       // Clear card detect change flag

          vTaskDelay(pdMS_TO_TICKS(40)); // Switch debounce delay
          switch (SD_CD) {
               case 0: // Card inserted
                    USB_PutQueue("SD card inserted\r\n");
                    SD_Card_Initialize();
                    InitFat32();
                    break;
               case 1: // Card removed
                    USB_PutQueue("SD card removed\r\n");
                    CardOK = FALSE;
                    FatInit = FALSE;
                    break;

          }
          MR15_bit = 1;       // Unmask interrupts
     }

}

/**************************************************************************************************
  SD card task
**************************************************************************************************/
//
void SDcardTask(void *pvParameters) {

     //init_timerpool();
     SD_Card_Initialize();
     InitFat32();
     RTC_Initialize();
     
     // Create a queue for SD card commands
     SDcardQueueHandle = xQueueCreate(MSG_QUEUE_LENGTH, sizeof(TSDcardMessage));
     vQueueAddToRegistry(SDcardQueueHandle,"SD Cmd Queue");

     while (1) {

          SDcardDetect();                    // Check SD card detect switch
          vTaskDelay(pdMS_TO_TICKS(50));     // every 50 ms
     }
}
//

/**************************************************************************************************
  End of File
**************************************************************************************************/