/**************************************************************************************************
  USART3_Driver3.c

  Copyright (C) 2015  Val Gretchev

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************************************

  Development Board: EasyMx PRO v7 for STM32 ARM
  Microcontroller: STM32F407VGT6
  Oscillator: external 25 MHz crystal
  MPU Frequency: 120 MHz from PLL

  MPU USART3 is directly connected to mikroBus slot 1 on RX and TX pins.

  Pin        Function   Slot 1  Switch  Signal
  (55) PD8   USART3_TX  TX      None    N/A
  (56) PD9   USART3_RX  RX      None    N/A

  Note: USART3 is dedicated to slot 1 use.


  For SYSCLK = 120 MHz and APB1 Prescaler = 4 the peripheral clock for APB1 = 30 MHz
  USART3 is on APB1 BUS and the peripheral clock is 30 MHz
  USARTDIV = Fck / 16 * baud = 30,000,000 / (16 * 9600) = 0d195.3125
  DIV_Mantissa = 0d195 = 0xC3
  DIV_Fraction = 16 * 0d0.3125 = 0d5 or 0x5
  Concatenate DIV Mantissa and DIV_Fraction,
  Therefore USART3_BRR = 0xC35
*/
/**************************************************************************************************
  USART3_Driver.c
**************************************************************************************************/
// Transparent mode is not dependent on control characters to end a transmission or reception.
// However, the user must know the length of the transmit message and the expected length of
// the received response and set variables appropriately. XmtByteCntr will hold the length of
// the transmit packet. The interrupt routine decrements this variable and is done when it reaches
// zero. U3RcvBufPtr will determine when the expected bytes of the reply packet have been received.
// Transparent mode is used when the WiFiESP is put is pass through mode.

// Include Files
#include <stdint.h>
#include <stddef.h>
#include "FreeRTOSConfig.h"
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"
#include "timers.h"
#include "projdefs.h"
#include "portmacro.h"

#include "Common.h"
#include "USART3_driver.h"
#include "Buffers.h"
#include "USB_Driver3.h"
//#include "MyTimers.h"
//#include "RTC.h"




struct AMessage *pxOutMessage, OutMessage;
struct AMessage *pxInMessage,  InMessage;

// Variables
char Transparent = 0;   // 0=Normal mode, 1=Transparent mode
unsigned RcvByteCntr;   // Not used
unsigned XmtByteCntr;   // Is set to the number of characters in a packet for Transparent mode
//struct AMessage *pxMessage;

// Queues
QueueHandle_t U3RcvByteQueueHandle;    // Queue of receive bytes
QueueHandle_t U3XmtPointerQueue;       // Queue of transmit message pointers

// Semaphores
SemaphoreHandle_t U3XmtSemHandle;      // Semaphore indicating transmitter is idle

extern QueueHandle_t vWiFiESPPointerQueue;


// Buffers
char U3RcvBuf[U3_RCV_BUF_SIZE];   // Receive messages are assembled in this buffer
unsigned U3RcvBufPtr = 0;         // Pointer to the next free byte in U3RcvBuf
char *U3XmtBuf;                   // Pointer to the current message being transmitted
unsigned U3XmtBufPtr;             // Points to next byte to be sent from transmit buffer
char U3BigBuf[U3_XMT_BUF_SIZE];   // Long transmit messages are assembled in this buffer


// The receive message routine has 2 states
tPointerFcn U3_RxFuncPtr = USART3_Rx1; // Start in State 1



/**************************************************************************************************
  USART3 interrupt handler
**************************************************************************************************/
void USART3_Interrupt() iv IVT_INT_USART3 ics ICS_AUTO {
     char R_data;        // Receive data byte
     char T_data;        // Transmit data byte
     BaseType_t xHigherPriorityTaskWoken = pdFALSE;

     if ((RXNE_USART3_SR_bit) /* || (ORE_bit) */ ) {
          R_data = USART3_DR;                // Fetch a byte from UART hardware
          xQueueSendFromISR(U3RcvByteQueueHandle,&R_data,&xHigherPriorityTaskWoken);
     }

     if ((TXE_USART3_SR_bit)&&(TXEIE_USART3_CR1_bit)) {
          if (Transparent == 0) {
              T_data = U3XmtBuf[U3XmtBufPtr++];  // Get a byte from the transmit buffer
              if (T_data == NUL) {               // If it is the end of message marker
                  TXEIE_USART3_CR1_bit = 0;      // Disable transmitter interrupts
                  xSemaphoreGiveFromISR(U3XmtSemHandle,&xHigherPriorityTaskWoken);
              } else {
                  USART3_DR = T_data;            // Transfer data byte to transmit register
              }
          } else { // XmtByteCounter must be set up by the application
              //if (XmtByteCntr == 0) {            // Alternate
              if (U3XmtBufPtr >= XmtByteCntr) {  // Use this instead of alternate
                  TXEIE_USART3_CR1_bit = 0;      // Disable transmitter interrupts
                  xSemaphoreGiveFromISR(U3XmtSemHandle,&xHigherPriorityTaskWoken);
              } else {
                  T_data = U3XmtBuf[U3XmtBufPtr++];   // Get a byte from the transmit buffer
                  USART3_DR = T_data;                 // Transfer data byte to transmit register
                  //XmtByteCntr--;                      // Alternate
              }
          }
     }
     portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}

/**************************************************************************************************
  This routine sets the baud rate.
**************************************************************************************************/
void USART3_SetBaud(void) {

     // Use MikroE library. UART library must be check marked
     
     // Configure USART3 for 115200 baud, 8-bits, no parity, 1 stop bit on pins PD8 (TX) and PD9 (RX)
     UART3_Init_Advanced(115200, _UART_8_BIT_DATA, _UART_NOPARITY, _UART_ONE_STOPBIT, &_GPIO_MODULE_USART3_PD89);
     
     // Configure USART3 for 9600 baud, 8-bits, no parity, 1 stop bit on pins PD8 (TX) and PD9 (RX)
     //UART3_Init_Advanced(9600, _UART_8_BIT_DATA, _UART_NOPARITY, _UART_ONE_STOPBIT, &_GPIO_MODULE_USART3_PD89);

     NVIC_IntEnable(IVT_INT_USART3); // Set interrupt vector
     NVIC_SetIntPriority(IVT_INT_USART3, _NVIC_INT_PRIORITY_LVL4);

}

/**************************************************************************************************
  This routine initializes the receive queue.
**************************************************************************************************/
void USART3_RxQueueInit(void) {
     U3RcvByteQueueHandle = xQueueCreate(U3_RCV_BYTE_Q_LENGTH, sizeof(char));
     if (U3RcvByteQueueHandle == NULL) {
         USB_PutQueue("Error: Could not create USART3 receive byte queue\r\n");
     } else {
         USB_PutQueue("Diag: USART3 receive byte queue created successfully\r\n");
         vQueueAddToRegistry(U3RcvByteQueueHandle, "U3 RcvByte Queue");
     }
}

/**************************************************************************************************
  This routine starts the receiver.
**************************************************************************************************/
void USART3_StartRx(void) {
     RXNEIE_USART3_CR1_bit = 1;        // Interrupt when receive buffer is full (RXNE=1)
}

/**************************************************************************************************
  USART3 receiver initialization
**************************************************************************************************/
void USART3_RcvInit(void) {
     USART3_SetBaud();
     USART3_RxQueueInit();
     USART3_StartRx();
}

/**************************************************************************************************
  UART_Rx State Machine. Call this function in the main loop.
**************************************************************************************************/
void USART3_Rx(void) {
     U3_RxFuncPtr();    // Use the function pointer to go to a state
}

/**************************************************************************************************
  USART3 receive subroutine
**************************************************************************************************/
void USART3_Rx1(void) {
     char RxChar;

     xQueueReceive(U3RcvByteQueueHandle, &RxChar, 0xFFFFFFFFU); // Block here

     if (Transparent == 0) { // Do this code for normal mode
         switch (RxChar) {
             case NUL:       // Discard NUL character
                 break;

             case '\r':      // Use return character
                 U3RcvBuf[U3RcvBufPtr++] = RxChar;    // Put received character into buffer
                 break;

             case '>':       // Prompt character is received without CRLF
             U3RcvBuf[U3RcvBufPtr++] = RxChar;   // Put received character into buffer
             U3RcvBuf[U3RcvBufPtr++] = '\r';
             U3RcvBuf[U3RcvBufPtr++] = '\n';
             U3RcvBuf[U3RcvBufPtr++] = 0;
             U3RcvBufPtr = 0;                    // Point to start of buffer
             U3_RxFuncPtr = USART3_Rx2;          // Go to state 2
             break;

             case '\n':      // Use new line character
                 U3RcvBuf[U3RcvBufPtr++] = RxChar;    // Put received character into buffer
                 U3RcvBuf[U3RcvBufPtr] = 0;           // End marker
                 U3RcvBufPtr = 0;                     // Point to start of buffer
                 U3_RxFuncPtr = USART3_Rx2;           // Go to state 2
                 break;


             default:
                 //if ((RxChar > 31) && (RxChar < 127)) {       // Use only ASCII set
                     if (U3RcvBufPtr < (U3_RCV_BUF_SIZE-4)) { // Prevents buffer overrun
                         U3RcvBuf[U3RcvBufPtr++] = RxChar;    // Put character in buffer
                     }                                        // Stay in this state
                 //}
                 break;
         }
     } else {    // Do this code for transparent mode
         if (U3RcvBufPtr < (U3_RCV_BUF_SIZE)) {  // Prevents buffer overrun
             U3RcvBuf[U3RcvBufPtr++] = RxChar;   // Put character in buffer
         }                                       // Stay in this state
     }
}

/**************************************************************************************************
  UART_Rx State 2: Execute the command indicated by the first byte in the buffer then return to
  State 1.
**************************************************************************************************/
void USART3_Rx2(void) {
     BaseType_t rtnCode;
     unsigned MsgLen;
     char *BufNum;

     BufNum = GetBuffer();             // Get a free buffer
     //USB_PutQueue(BufNum);           // Send copy to USB (Done in WiFi_ESP_Driver)

     //
     if (U3RcvBuf[0] != '\r') {        // Throw away blank lines
         MsgLen = strlen(U3RcvBuf);
         if (MsgLen <= BUFSIZE) {      // Make sure the message will fit
             strcpy(BufNum,U3RcvBuf);  // Make a copy of the received message
             rtnCode = xQueueSend(vWiFiESPPointerQueue, &BufNum, 0); // No blocking
         } else {
             rtnCode = xQueueSend(vWiFiESPPointerQueue, &U3RcvBuf, 0); // No blocking
         }
         // Diagnostic
         if (rtnCode == pdFAIL)
             USB_PutQueue("Error: Failed to enter message pointer into WiFiESP queue\r\n");
     }
     //

     U3_RxFuncPtr = USART3_Rx1;    // Go back to state 1
}

/**************************************************************************************************
  USART3 receive task
**************************************************************************************************/
//
void USART3_RcvTask(void *pvParameters) {

     USART3_RcvInit();

     while (1) {
          USART3_Rx();    // Check for incoming USART3 messages

          // Delay
          //vTaskDelay(pdMS_TO_TICKS(10)); // Delay for 10ms
     }
}
//
/*************************************************************************************************/










/**************************************************************************************************
  USART3 put queue text.
**************************************************************************************************/
void USART3_PutQueueText(char *Msg) {
     BaseType_t rtnCode;
     //char *BufNum;

     OutMessage.MsgLen = strlen(Msg);  // Length of message
     OutMessage.TheMessage = Msg;      // Pointer to the message

     /*
     BufNum = GetBuffer();
     sprintf(BufNum,"Message length: %d\r\n",OutMessage.MsgLen);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"Message: %d\r\n",OutMessage.TheMessage);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"Message: %s\r\n",OutMessage.TheMessage);
     USB_PutQueue(BufNum);
     */

     // Put the structure into the queue
     rtnCode = xQueueSend(U3XmtPointerQueue, &OutMessage, 0); // No blocking

     if (rtnCode == pdFAIL)
          USB_PutQueue("Error: Failed to enter structure into USART3 queue\r\n");

}

/**************************************************************************************************
  USART3 put queue binary.
**************************************************************************************************/
void USART3_PutQueueBinary(char *Msg, unsigned Len) {
     BaseType_t rtnCode;

     OutMessage.MsgLen = Len;          // Length of message
     OutMessage.TheMessage = Msg;      // Pointer to message

     // Put the structure into the queue
     rtnCode = xQueueSend(U3XmtPointerQueue, &OutMessage, 0); // No blocking

     if (rtnCode == pdFAIL)
          USB_PutQueue("Error: Failed to enter structure into USART3 queue\r\n");

}

/**************************************************************************************************
  This function waits for USART3 to be idle before proceeding. In the next step the function
  waits indefinitely for an entry in the queue. Then it proceeds to extract the message length
  and the pointer to the message and starts the transmitter.
**************************************************************************************************/
void USART3_Tx(void) {
     BaseType_t rtnCode; // Function return code
     //char *BufNum;

     // If the transmitter is busy, wait here
     xSemaphoreTake(U3XmtSemHandle, portMAX_DELAY); // Block here

     // If there are no messages to send, wait here
     rtnCode = xQueueReceive(U3XmtPointerQueue, &InMessage, 0xFFFFFFFFU); // Block here

     U3XmtBuf = InMessage.TheMessage;    // Extract the message pointer
     XmtByteCntr = InMessage.MsgLen;     // Extract the message length
     U3XmtBufPtr = 0;             // Set character counter to start of text message
     TXEIE_USART3_CR1_bit = 1;    // Interrupt when transmit buffer is empty (TXE=1)

     // Un comment char *BufNum when using below diagnostics.
     /*
     BufNum = GetBuffer();
     sprintf(BufNum,"Message length: %d\r\n",InMessage.MsgLen);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"Message: %d\r\n",InMessage.TheMessage);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"Message: %s\r\n",InMessage.TheMessage);
     USB_PutQueue(BufNum);
     */

}




/**************************************************************************************************
  USART3 transmit task
**************************************************************************************************/
//
void USART3_XmtTask(void *pvParameters) {

     pxOutMessage = &OutMessage;
     pxInMessage  = &InMessage;

     U3XmtSemHandle = xSemaphoreCreateBinary(); // Semaphore indicating transmitter is idle
     if (U3XmtSemHandle == NULL) {
         USB_PutQueue("Error: Could not create USART3 transmit semaphore\r\n");
     } else {
         USB_PutQueue("Diag: USART3 transmit semaphore created successfully\r\n");
         vQueueAddToRegistry(U3XmtSemHandle, "USART3 transmit Semaphore");
     }

     //U3XmtPointerQueue = xQueueCreate(64, sizeof(char *)); // Queue of message pointers
     U3XmtPointerQueue = xQueueCreate(64, sizeof(struct AMessage));
     if (U3XmtPointerQueue == NUL) {
         USB_PutQueue("Error: Could not create USART3 transmit pointer queue\r\n");
     } else {
         USB_PutQueue("Diag: USART3 transmit pointer queue created successfully\r\n");
         vQueueAddToRegistry(U3XmtPointerQueue, "USART3 pointer queue");
     }


     xSemaphoreGive(U3XmtSemHandle); // This gets us past the semaphore block initially
     USB_PutQueue("Diag: USART3 task started\r\n");

     while (1) {
          USART3_Tx(); // Check for outgoing USART3 messages

          // Delay not required
          //vTaskDelay(pdMS_TO_TICKS(50)); // Delay for 50ms
     }
}
//






/**************************************************************************************************
  Diagnostic: Display USART3 registers
**************************************************************************************************/
void ShowUSART3Regs(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"USART3_CR1 register:  %08LX\r\n",USART3_CR1);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"USART3_CR2 register:  %08LX\r\n",USART3_CR2);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"USART3_CR3 register:  %08LX\r\n",USART3_CR3);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"USART3_BRR register:  %08LX\r\n",USART3_BRR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"USART3_SR register:   %08LX\r\n",USART3_SR);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"GPIOD_MODER register: %08LX\r\n",GPIOD_MODER);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"GPIOD_AFRH register:  %08LX\r\n",GPIOD_AFRH);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"GPIOD_AFRL register:  %08LX\r\n",GPIOD_AFRL);
     USB_PutQueue(BufNum);
}

/**************************************************************************************************
  End of File
**************************************************************************************************/